import math

from pan3d.utils.common import RenderingSettingsBasic
from pan3d.utils.constants import XYZ
from pan3d.utils.convert import max_str_length
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class GlobeRenderingSettings(RenderingSettingsBasic):
    def __init__(self, source, update_rendering, **kwargs):
        super().__init__(source, update_rendering, **kwargs)

        self.source = source

        with self.content:
            v3.VDivider()
            v3.VSelect(
                placeholder="Globe Texture",
                prepend_inner_icon="mdi-earth",
                v_model=("texture", self.state.textures[0]),
                items=("textures", []),
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
            )
            v3.VDivider()
            v3.VSelect(
                placeholder="Data Representation",
                prepend_inner_icon=(
                    "['mdi-dots-triangle', 'mdi-triangle-outline','mdi-triangle'][representation]",
                ),
                v_model=("representation", 2),
                items=(
                    "representations",
                    [
                        {"title": "Surface", "value": 2},
                        {"title": "Wireframe", "value": 1},
                        {"title": "Points", "value": 0},
                    ],
                ),
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
            )
            v3.VDivider()
            with v3.VTooltip(
                v_if="representation !== 2",
                text=(
                    "`${representation ? 'Line Size' : 'Point Size' }: ${cell_size} - Shadow: ${render_shadow ? 'On': 'Off'}`",
                ),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            classes="pr-3 ml-3",
                            prepend_icon="mdi-format-line-weight",
                            v_model=("cell_size", 1),
                            min=1,
                            max=100,
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                            click_prepend="render_shadow = !render_shadow",
                        )
            with v3.VTooltip(
                v_if="representation === 2",
                text=("`Opacity: ${opacity.toFixed(2)}`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            classes="pr-3 ml-3",
                            prepend_icon="mdi-circle-opacity",
                            v_model=("opacity", 1.0),
                            min=0.0,
                            max=1.0,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )

            with v3.VTooltip(
                text=("`Bump Radius: ${bump_radius}`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            classes="pr-3 ml-3",
                            prepend_icon="mdi-signal-distance-variant",
                            v_model=("bump_radius", 10),
                            min=10,
                            max=1000,
                            step=10,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )

            v3.VDivider()

            # X crop/cut
            with v3.VTooltip(
                v_if="axis_names?.[0]",
                text=(
                    "`${axis_names[0]}: [${dataset_bounds[0]}, ${dataset_bounds[1]}] ${slice_x_type ==='range' ? ('(' + slice_x_range.map((v,i) => v+1).concat(slice_x_step).join(', ') + ')'): slice_x_cut}`",
                ),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex",
                        v_if="axis_names?.[0]",
                        v_bind="props",
                    ):
                        v3.VRangeSlider(
                            v_if="slice_x_type === 'range'",
                            prepend_icon="mdi-axis-x-arrow",
                            v_model=("slice_x_range", None),
                            min=("slice_extents[axis_names[0]][0]",),
                            max=("slice_extents[axis_names[0]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VSlider(
                            v_else=True,
                            prepend_icon="mdi-axis-x-arrow",
                            v_model=("slice_x_cut", 0),
                            min=("slice_extents[axis_names[0]][0]",),
                            max=("slice_extents[axis_names[0]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VCheckbox(
                            v_model=("slice_x_type", "range"),
                            true_value="range",
                            false_value="cut",
                            true_icon="mdi-crop",
                            false_icon="mdi-box-cutter",
                            hide_details=True,
                            density="compact",
                            size="sm",
                            classes="mx-2",
                        )

            # Y crop/cut
            with v3.VTooltip(
                v_if="axis_names?.[1]",
                text=(
                    "`${axis_names[1]}: [${dataset_bounds[2]}, ${dataset_bounds[3]}] ${slice_y_type ==='range' ? ('(' + slice_y_range.map((v,i) => v+1).join(', ') + ', 1)'): slice_y_cut}`",
                ),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex",
                        v_if="axis_names?.[1]",
                        v_bind="props",
                    ):
                        v3.VRangeSlider(
                            v_if="slice_y_type === 'range'",
                            prepend_icon="mdi-axis-y-arrow",
                            v_model=("slice_y_range", None),
                            min=("slice_extents[axis_names[1]][0]",),
                            max=("slice_extents[axis_names[1]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VSlider(
                            v_else=True,
                            prepend_icon="mdi-axis-y-arrow",
                            v_model=("slice_y_cut", 0),
                            min=("slice_extents[axis_names[1]][0]",),
                            max=("slice_extents[axis_names[1]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VCheckbox(
                            v_model=("slice_y_type", "range"),
                            true_value="range",
                            false_value="cut",
                            true_icon="mdi-crop",
                            false_icon="mdi-box-cutter",
                            hide_details=True,
                            density="compact",
                            size="sm",
                            classes="mx-2",
                        )

            # Z crop/cut
            with v3.VTooltip(
                v_if="axis_names?.[2]",
                text=(
                    "`${axis_names[2]}: [${dataset_bounds[4]}, ${dataset_bounds[5]}] ${slice_z_type ==='range' ? ('(' + slice_z_range.map((v,i) => v+1).join(', ') + ', 1)'): slice_z_cut}`",
                ),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex",
                        v_bind="props",
                    ):
                        v3.VRangeSlider(
                            v_if="slice_z_type === 'range'",
                            prepend_icon="mdi-axis-z-arrow",
                            v_model=("slice_z_range", None),
                            min=("slice_extents[axis_names[2]][0]",),
                            max=("slice_extents[axis_names[2]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VSlider(
                            v_else=True,
                            prepend_icon="mdi-axis-z-arrow",
                            v_model=("slice_z_cut", 0),
                            min=("slice_extents[axis_names[2]][0]",),
                            max=("slice_extents[axis_names[2]][1]",),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
                        v3.VCheckbox(
                            v_model=("slice_z_type", "range"),
                            true_value="range",
                            false_value="cut",
                            true_icon="mdi-crop",
                            false_icon="mdi-box-cutter",
                            hide_details=True,
                            density="compact",
                            size="sm",
                            classes="mx-2",
                        )
            v3.VDivider()

            # Slice steps
            with v3.VTooltip(text="Level Of Details / Slice stepping"):
                with html.Template(v_slot_activator="{ props }"):
                    with v3.VRow(
                        v_bind="props",
                        no_gutter=True,
                        classes="align-center my-0 mx-0 border-b-thin",
                    ):
                        v3.VIcon(
                            "mdi-stairs",
                            classes="ml-2 text-medium-emphasis",
                        )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[0]"):
                            v3.VTextField(
                                v_model_number=("slice_x_step", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=['min="1"'],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[1]"):
                            v3.VTextField(
                                v_model_number=("slice_y_step", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=['min="1"'],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[2]"):
                            v3.VTextField(
                                v_model_number=("slice_z_step", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=['min="1"'],
                                type="number",
                            )
            # Time slider
            with v3.VTooltip(
                v_if="slice_t_max > 0",
                text=("`time: ${t_labels[slice_t]} (${slice_t+1}/${slice_t_max+1})`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            prepend_icon="mdi-clock-outline",
                            v_model=("slice_t", 0),
                            min=0,
                            max=("slice_t_max", 0),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
            v3.VDivider()
            v3.VBtn(
                "Update 3D view",
                block=True,
                classes="text-none",
                flat=True,
                density="compact",
                rounded=0,
                disabled=("data_arrays.length === 0",),
                color=("dirty_data && data_arrays.length ? 'primary': undefined",),
                click=(update_rendering, "[true]"),
            )

    def update_from_source(self, source=None):
        if source is None:
            return

        with self.state as state:
            state.data_arrays_available = source.available_arrays
            state.data_arrays = source.arrays
            state.color_by = None
            state.axis_names = [source.x, source.y, source.z]
            state.slice_extents = source.slice_extents
            slices = source.slices
            for axis in XYZ:
                # default
                axis_extent = state.slice_extents.get(getattr(source, axis))
                state[f"slice_{axis}_range"] = axis_extent
                state[f"slice_{axis}_cut"] = 0
                state[f"slice_{axis}_step"] = 1
                state[f"slice_{axis}_type"] = "range"

                # use slice info if available
                axis_slice = slices.get(getattr(source, axis))
                if axis_slice is not None:
                    if isinstance(axis_slice, int):
                        # cut
                        state[f"slice_{axis}_cut"] = axis_slice
                        state[f"slice_{axis}_type"] = "cut"
                    else:
                        # range
                        state[f"slice_{axis}_range"] = [
                            axis_slice[0],
                            axis_slice[1] - 1,
                        ]  # end is inclusive
                        state[f"slice_{axis}_step"] = axis_slice[2]

            # Update time
            state.slice_t = source.t_index
            state.slice_t_max = source.t_size - 1
            state.t_labels = source.t_labels
            state.max_time_width = math.ceil(0.58 * max_str_length(state.t_labels))
            if state.slice_t_max > 0:
                state.max_time_index_width = math.ceil(
                    0.6 + (math.log10(state.slice_t_max + 1) + 1) * 2 * 0.58
                )



import hashlib
import sys

from trame.decorators import TrameApp, change
from trame.widgets import html, plotly
from trame.widgets import vuetify3 as v3

try:
    import logging

    import xcdat  # noqa: F401

    logging.getLogger().setLevel(logging.CRITICAL + 1)
except ModuleNotFoundError as e:
    print(
        f"""
        Error occurred while importing xcdat required for Analytics Explorer: {e}
        Please make sure xcdat is installed properly using conda -- `conda install -c conda-forge xcdat`
        You may want to create a new conda environment an reinstall pan3d within along with xcdat.
        """
    )
    sys.exit(1)
from enum import Enum

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from pan3d.xarray.algorithm import to_isel


class PlotTypes(Enum):
    ZONAL = (0,)
    ZONALTIME = (1,)
    GLOBAL = (2,)
    TEMPORAL = (3,)


plot_options = {
    PlotTypes.ZONAL: "Zonal Average",
    PlotTypes.ZONALTIME: "Zonal Average w/ Time",
    PlotTypes.GLOBAL: "Global Average",
    PlotTypes.TEMPORAL: "Temporal Average",
}


class GroupBy(Enum):
    HOUR = 0
    DAY = 1
    MONTH = 2
    SEASON = 3
    YEAR = 4
    NONE = 5


group_options = {
    GroupBy.HOUR: "Hour",
    GroupBy.DAY: "Day",
    GroupBy.MONTH: "Month",
    GroupBy.SEASON: "Season",
    GroupBy.YEAR: "Year",
    GroupBy.NONE: "None",
}

zonal_axes = {
    "Longitude": "X",
    "Latitude": "Y",
}


class CacheEntry:
    def __init__(self, data, figure):
        self.data = data
        self.figure = figure


@TrameApp()
class Plotting(v3.VCard):
    def __init__(
        self,
        source=None,
        **kwargs,
    ):
        super().__init__(
            rounded=True,
            classes="h-100",
            **kwargs,
        )
        self.source = source
        self.spatial_cache = {}
        self.temporal_cache = {}
        self.zonal_cache = {}

        # State variables controlling the UI for various types of plots
        self.state.figure_height = 50
        self.state.show_group_by = False
        self.state.show_temporal_slider = False
        self.state.show_zonal_axis = True
        self.state.show_update_button = False

        with self:
            with v3.VCardTitle():
                with html.Div(classes="d-flex"):
                    v3.VSelect(
                        label="Plot Type",
                        v_model=("active_plot", plot_options.get(PlotTypes.ZONAL)),
                        items=("plots", list(plot_options.values())),
                        hide_details=True,
                        variant="outlined",
                        density="compact",
                    )
                v3.VDivider(classes="my-2")
                with html.Div(classes="d-flex"):
                    v3.VSelect(
                        v_show="show_zonal_axis",
                        label="Zonal Axis",
                        v_model=("zonal_axis", next(iter(zonal_axes.keys()))),
                        items=("axes", list(zonal_axes.keys())),
                        hide_details=True,
                        variant="outlined",
                        density="compact",
                    )
                    v3.VSelect(
                        v_show="show_group_by",
                        label="Group By",
                        v_model=("group_by", group_options.get(GroupBy.YEAR)),
                        items=("groups", list(group_options.values())),
                        hide_details=True,
                        variant="outlined",
                        density="compact",
                    )
                with html.Div(
                    v_if="show_temporal_slider",
                ):
                    html.Div("Temporal Slice Selection")
                    v3.VSlider(
                        v_model=("temporal_slice", 0),
                        min=0,
                        max=("time_groups", 0),
                        step=1,
                        hide_details=True,
                        density="compact",
                    )
                v3.VDivider(classes="my-2")
                with html.Div(
                    v_if="show_update_button", classes="text-center align-center"
                ):
                    v3.VBtn("Update Plots", click=self.update_plot)
            with v3.VCardText(style=("`height: ${figure_height}%;`",)):
                figure = plotly.Figure(
                    display_logo=True,
                    display_mode_bar=True,
                )
                self.ctrl.figure_update = figure.update

    def expose_plot_specific_config(self):
        """
        Toggle visibility of components controlling plots based on chosen type
        """
        state = self.state
        plot_type = state.active_plot

        # Default values
        config = {
            "figure_height": 50,
            "show_zonal_axis": False,
            "show_group_by": True,
            "show_temporal_slider": False,
            "show_update_button": True,
        }

        # Overrides for each plot type
        plot_overrides = {
            plot_options.get(PlotTypes.ZONAL): {
                "show_zonal_axis": True,
                "show_group_by": False,
                "show_update_button": False,
            },
            plot_options.get(PlotTypes.ZONALTIME): {
                "figure_height": 80,
                "show_zonal_axis": True,
            },
            plot_options.get(PlotTypes.GLOBAL): {
                "show_zonal_axis": False,
            },
            plot_options.get(PlotTypes.TEMPORAL): {
                "show_temporal_slider": True,
            },
        }

        # Merge overrides only if the plot type exists in the dictionary
        config.update(plot_overrides.get(plot_type, {}))

        # Update state in a single call
        state.update(config)

    def get_key(self, selection):
        return hashlib.sha256(repr(selection).encode()).hexdigest()

    def get_selection_criteria(self, full_temporal=False):
        """
        Get the xarray slicing criteria based on current selection of data
        """
        slices = self.source.slices
        (x, y, z, t) = (self.source.x, self.source.y, self.source.z, self.source.t)
        if full_temporal:
            return to_isel(slices, x, y, z)
        return to_isel(slices, x, y, z, t)

    def apply_spatial_average(self, axis=None):
        """
        Calculate spatial average of data for the current selected time
        """
        if axis is None:
            axis = ["X"]
        ds = self.source.input
        active_var = self.state.color_by
        select = self.get_selection_criteria()
        return ds.isel(select).spatial.average(active_var, axis=axis)

    def apply_spatial_average_full_temporal(self, axis=None):
        """
        Calculate spatial average for data for full temporal resoulution
        """
        if axis is None:
            axis = ["X"]

        ds = self.source.input
        active_var = self.state.color_by

        select = self.get_selection_criteria(full_temporal=True)
        average = (
            ds.isel(select).spatial.average(active_var, axis)
            if axis is not None
            else ds.isel(select).spatial.average(active_var)
        )

        group_by = self.state.group_by
        if group_by == group_options.get(GroupBy.NONE):
            return average
        return average.temporal.group_average(
            active_var, freq=group_by.lower(), weighted=True
        )

    def apply_temporal_average(self, active_var):
        """
        Calculate time based average for data
        """
        ds = self.source.input
        select = self.get_selection_criteria(full_temporal=True)
        group_by = self.state.group_by
        if group_by == group_options.get(GroupBy.NONE):
            return ds.isel(select).temporal.average(active_var, weighted=True)
        return ds.isel(select).temporal.group_average(
            active_var, freq=group_by.lower(), weighted=True
        )

    def generate_plot(self):
        state = self.state
        active_var = self.state.color_by
        (x, y, _, t) = (self.source.x, self.source.y, self.source.z, self.source.t)
        if active_var is None:
            return None

        plot_type = state.active_plot
        zonal_axis = state.zonal_axis
        axis = x if zonal_axes.get(zonal_axis) == "X" else y
        if plot_type == plot_options.get(PlotTypes.ZONAL):
            return self.zonal_average(active_var, axis)
        if plot_type == plot_options.get(PlotTypes.ZONALTIME):
            return self.zonal_with_time(active_var, axis, t)
        if plot_type == plot_options.get(PlotTypes.GLOBAL):
            return self.global_full_temporal(active_var, t)
        if plot_type == plot_options.get(PlotTypes.TEMPORAL):
            return self.temporal_average(active_var, x, y, t)
        return None

    def zonal_average(self, active_var, axis):
        """
        Get a plotly figure for the zonal average for current sptio-temporal selection.
        Average is calculated over a certain specified spatial dimension (Longitude or Latitude).
        """
        data = self.apply_spatial_average(axis=zonal_axes.get(self.state.zonal_axis))
        to_plot = data[active_var]
        dim = to_plot.dims[0]
        plot = px.line(x=to_plot[dim].to_numpy(), y=to_plot.to_numpy())

        var_long_name = data[active_var].attrs.get("long_name", active_var)
        var_units = data[active_var].attrs.get("units", "")
        x_axis_name = data[dim].attrs.get("long_name", dim)
        plot.update_layout(
            title=f'Zonal Average for {active_var} "{var_long_name}" over {axis} (unit: {var_units})',
            xaxis_title=x_axis_name,
            yaxis_title=var_long_name,
        )
        return plot

    def zonal_with_time(self, active_var, axis, t):
        """
        Get a plotly figure for the zonal average along with current and full temporal selection.
        Average is calculated over a certain specified spatial dimension (Longitude or Latitude).
        """
        data_t = self.apply_spatial_average(axis=zonal_axes.get(self.state.zonal_axis))
        data = self.apply_spatial_average_full_temporal(
            axis=zonal_axes.get(self.state.zonal_axis)
        )

        var_long_name = data[active_var].attrs.get("long_name", active_var)
        var_units = data[active_var].attrs.get("units", "")
        figure = make_subplots(
            rows=2,
            cols=1,
            subplot_titles=(
                f'Zonal Average for {active_var} "{var_long_name}" at {t} = {self.state.slice_t} (unit: {var_units})',
                f"Zonal Average for {var_long_name}",
            ),
        )
        to_plot = data_t[active_var]
        dim = to_plot.dims[0]
        x_axis_name = data[dim].attrs.get("long_name", dim)
        plot = px.line(x=to_plot[dim], y=to_plot)
        for trace in plot.data:
            figure.add_trace(trace, row=1, col=1)
        figure.update_xaxes(title_text=x_axis_name, row=1, col=1)
        figure.update_yaxes(title_text=var_long_name, row=1, col=1)

        taxis = self.get_time_labels(data[t])
        plot = px.imshow(data[active_var], y=taxis)
        plot.update_layout(coloraxis_colorbar={"orientation": "h"})
        for trace in plot.data:
            figure.add_trace(trace, row=2, col=1)
        figure.update_xaxes(title_text=x_axis_name, row=2, col=1)
        figure.update_yaxes(title_text=t, row=2, col=1)

        return figure

    def global_full_temporal(self, active_var, t):
        """
        Get a plotly figure for the global average for all data with full temporal resolution.
        Data from spatial dimension is averaged yielding a single quantity with tempoal dimension.
        """
        data = self.apply_spatial_average_full_temporal(axis=None)
        time = self.get_time_labels(data[t])
        plot = px.line(x=time, y=data[active_var])
        # plot.update_layout(title_text=f"Global Average for {active_var}")

        var_long_name = data[active_var].attrs.get("long_name", active_var)
        var_units = data[active_var].attrs.get("units", "")
        x_axis_name = data[t].attrs.get("long_name", t)
        plot.update_layout(
            title=f'Global average for {active_var} "{var_long_name}" (unit: {var_units})',
            xaxis_title=x_axis_name,
            yaxis_title=var_long_name,
        )
        return plot

    def temporal_average(self, active_var, x, y, t):
        """
        Get a time based average of data, data in temporal domain in averaged keeping spatial dimensions same
        """
        data = self.apply_temporal_average(active_var)
        self.state.time_groups = len(data[t])
        slice = self.state.temporal_slice
        plot = px.imshow(data[active_var][slice])
        plot.update_yaxes(autorange="reversed")
        var_long_name = data[active_var].attrs.get("long_name", active_var)
        var_units = data[active_var].attrs.get("units", "")
        x_axis_name = data[x].attrs.get("long_name", x)
        y_axis_name = data[y].attrs.get("long_name", y)
        plot.update_layout(
            title=f'Temporal average for {active_var} "{var_long_name}" (unit: {var_units}) at time {slice}/{len(data[t])}',
            xaxis_title=x_axis_name,
            yaxis_title=y_axis_name,
            coloraxis_colorbar={"orientation": "h"},
        )

        return plot

    def get_time_labels(self, time_array):
        group_by = self.state.group_by
        if group_by == group_options.get(GroupBy.YEAR):
            return np.vectorize(lambda dt: f"{dt.year}")(time_array)
        if group_by == group_options.get(GroupBy.MONTH):
            return np.vectorize(lambda dt: f"{dt.month}-{dt.year}")(time_array)
        return np.vectorize(lambda dt: f"{dt.isoformat()}")(time_array)

    @change("temporal_slice")
    def on_change_group(self, **kwargs):
        active_var = self.state.color_by
        (x, y, _, t) = (self.source.x, self.source.y, self.source.z, self.source.t)
        if active_var is None:
            return
        plot = self.temporal_average(active_var, x, y, t)
        self.ctrl.figure_update(plot)

    @change(
        "zonal_axis",
        "slice_t",
        "slice_x_range",
        "slice_y_range",
        "slice_z_range",
    )
    def update_plot(self, **kwargs):
        figure = self.generate_plot()
        self.ctrl.figure_update(figure)

    @change("active_plot")
    def on_change_active_plot(self, **kwargs):
        self.expose_plot_specific_config()
        plot_type = self.state.active_plot
        if plot_type != plot_options.get(PlotTypes.ZONAL):
            self.ctrl.figure_update(go.Figure())
            return
        figure = self.generate_plot()
        self.ctrl.figure_update(figure)



import math

import numpy as np
from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.util import numpy_support, vtkConstants
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkFiltersCore import vtkAppendFilter


def ProcessPoint(point, radius, scale):
    theta = point[0]
    phi = 90 - point[1]
    rho = point[2] * scale + radius if point[2] != 0 else radius
    x = rho * math.sin(math.radians(phi)) * math.cos(math.radians(theta))
    y = rho * math.sin(math.radians(phi)) * math.sin(math.radians(theta))
    z = rho * math.cos(math.radians(phi))
    return [x, y, z]


class ProjectToSphere(VTKPythonAlgorithmBase):
    def __init__(self):
        super().__init__(
            nInputPorts=1, nOutputPorts=1, outputType="vtkUnstructuredGrid"
        )
        self.__Dims = -1
        self.isData = False
        self.radius = 6378
        self.scale = 1.0
        self._bump_radius = 10

    def SetDataLayer(self, isData_):
        if self.isData != isData_:
            self.isData = isData_
            self.Modified()

    def SetScalingFactor(self, scale_):
        if self.scale != float(scale_):
            self.scale = float(scale_)
            self.Modified()

    @property
    def bump_radius(self):
        return self._bump_radius

    @bump_radius.setter
    def bump_radius(self, v):
        if v != self._bump_radius:
            self._bump_radius = v
            self.Modified()

    def RequestData(self, request, inInfo, outInfo):
        inData = self.GetInputData(inInfo, 0, 0)
        outData = self.GetOutputData(outInfo, 0)
        if not inData.IsA("vtkUnstructuredGrid"):
            afilter = vtkAppendFilter()
            afilter.AddInputData(inData)
            afilter.Update()
            outData.DeepCopy(afilter.GetOutput())
        else:
            outData.DeepCopy(inData)

        outWrap = dsa.WrapDataObject(outData)
        try:
            inPoints = np.array(outWrap.Points)
            pRadius = (self.radius + self._bump_radius) if self.isData else self.radius
            outPoints = np.array(
                [ProcessPoint(x, pRadius, self.scale) for x in inPoints]
            )
        except Exception as e:
            print(e)
        _coords = numpy_support.numpy_to_vtk(
            outPoints, deep=True, array_type=vtkConstants.VTK_FLOAT
        )
        vtk_coords = vtkPoints()
        vtk_coords.SetData(_coords)
        outWrap.SetPoints(vtk_coords)

        return 1



import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkDataSetAttributes
from vtkmodules.vtkFiltersCore import (
    vtkAssignAttribute,
    vtkCellDataToPointData,
    vtkTriangleFilter,
)
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkFiltersModeling import (
    vtkBandedPolyDataContourFilter,
    vtkLoopSubdivisionFilter,
)

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleSwitch  # noqa: F401
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

from pan3d.ui.contour import ContourRenderingSettings
from pan3d.ui.vtk_view import Pan3DView
from pan3d.utils.common import ControlPanel, Explorer, SummaryToolbar
from pan3d.utils.convert import to_float
from pan3d.widgets.scalar_bar import ScalarBar
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import vuetify3 as v3


class ContourExplorer(Explorer):
    def __init__(self, xarray=None, source=None, server=None, local_rendering=None):
        super().__init__(xarray, source, server, local_rendering)

        if self.source is None:
            self.source = vtkXArrayRectilinearSource(
                input=self.xarray
            )  # To initialize the pipeline

        # setup
        self.last_field = None
        self.last_preset = None

        self._setup_vtk()
        self._build_ui()

    # -------------------------------------------------------------------------
    # VTK Setup
    # -------------------------------------------------------------------------

    def _setup_vtk(self):
        ds = self.source()

        self.renderer = vtkRenderer(background=(0.8, 0.8, 0.8))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        # Need explicit geometry extraction when used with WASM
        self.geometry = vtkDataSetSurfaceFilter(
            input_connection=self.source.output_port
        )
        self.triangle = vtkTriangleFilter(input_connection=self.geometry.output_port)
        self.cell2point = vtkCellDataToPointData(
            input_connection=self.triangle.output_port
        )
        self.refine = vtkLoopSubdivisionFilter(
            input_connection=self.cell2point.output_port, number_of_subdivisions=1
        )
        self.assign = vtkAssignAttribute(input_connection=self.refine.output_port)
        self.assign.Assign(
            None,
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        self.bands = vtkBandedPolyDataContourFilter(
            input_connection=self.assign.output_port,
            generate_contour_edges=1,
        )
        self.mapper = vtkPolyDataMapper(
            input_connection=self.bands.output_port,
            scalar_visibility=1,
            interpolate_scalars_before_mapping=1,
        )
        self.mapper.SetScalarModeToUsePointFieldData()
        self.actor = vtkActor(mapper=self.mapper)

        # contour lines
        self.mapper_lines = vtkPolyDataMapper(
            input_connection=self.bands.GetOutputPort(1),
        )
        self.actor_lines = vtkActor(mapper=self.mapper_lines)
        self.actor_lines.property.color = [0, 0, 0]
        self.actor_lines.property.line_width = 2

        self.renderer.AddActor(self.actor)
        self.renderer.AddActor(self.actor_lines)

        self.renderer.ResetCamera(ds.bounds)

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    # -------------------------------------------------------------------------
    # UI
    # -------------------------------------------------------------------------

    def _build_ui(self, **_):
        self.state.update(
            {
                "trame__title": "Contour Explorer",
                "import_pending": False,
                "control_expended": True,
                "axis_names": ["X", "Y", "Z"],
                "scale_x": 1,
                "scale_y": 1,
                "scale_z": 0.01,
            }
        )
        with VAppLayout(self.server, fill_height=True) as layout:
            self.ui = layout

            # 3D view
            Pan3DView(
                self.render_window,
                local_rendering=self.local_rendering,
                widgets=[self.widget],
            )

            # Scalar bar
            ScalarBar(
                ctx_name="scalar_bar",
                v_show="!control_expended",
                v_if="color_by",
            )

            # Save dialog
            with v3.VDialog(v_model=("show_save_dialog", False)):
                with v3.VCard(classes="mx-auto w-50"):
                    v3.VCardTitle("Save dataset to disk")
                    v3.VDivider()
                    with v3.VCardText():
                        v3.VTextField(
                            label="File path to save",
                            v_model=("save_dataset_path", ""),
                            hide_details=True,
                        )
                    with v3.VCardActions():
                        v3.VSpacer()
                        v3.VBtn(
                            "Save",
                            classes="text-none",
                            variant="flat",
                            color="primary",
                            click=(self.save_dataset, "[save_dataset_path]"),
                        )
                        v3.VBtn(
                            "Cancel",
                            classes="text-none",
                            variant="flat",
                            click="show_save_dialog=false",
                        )

            # Error messages
            v3.VAlert(
                v_if=("data_origin_error", False),
                border="start",
                max_width=700,
                rounded="lg",
                text=("data_origin_error", ""),
                title="Failed to load data",
                type="error",
                variant="tonal",
                style="position:absolute;bottom:1rem;right:1rem;",
            )

            # Summary toolbar
            SummaryToolbar(
                v_show="!control_expended",
                v_if="slice_t_max > 0",
            )

            # Control panel
            with ControlPanel(
                enable_data_selection=(self.xarray is None),
                source=self.source,
                toggle="control_expended",
                load_dataset=self.load_dataset,
                import_file_upload=self.import_file_upload,
                export_file_download=self.export_state,
                xr_update_info="xr_update_info",
                panel_label="Contour Explorer",
            ).ui_content:
                ContourRenderingSettings(
                    ctx_name="rendering",
                    source=self.source,
                    update_rendering=self.update_rendering,
                )

    def update_rendering(self, reset_camera=False):
        self.renderer.ResetCamera()

        if self.local_rendering:
            self.ctrl.view_update(push_camera=True)

        self.ctrl.view_reset_camera()

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("scale_x", "scale_y", "scale_z")
    def _on_scale_change(self, scale_x, scale_y, scale_z, **_):
        self.actor.SetScale(
            to_float(scale_x),
            to_float(scale_y),
            to_float(scale_z),
        )
        self.actor_lines.SetScale(
            to_float(scale_x),
            to_float(scale_y),
            to_float(scale_z),
        )
        self.renderer.ResetCamera()

        if self.local_rendering:
            self.ctrl.view_update(push_camera=True)

        self.ctrl.view_reset_camera()

    @change("color_by", "nb_contours", "color_min", "color_max")
    def _on_color_by_change(
        self, color_by, nb_contours, color_min, color_max, **kwargs
    ):
        self.assign.Assign(
            color_by,
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        self.bands.GenerateValues(nb_contours, [color_min, color_max])
        super()._on_color_properties_change(**kwargs)


def main():
    app = ContourExplorer()
    app.server.start()


if __name__ == "__main__":
    main()



import warnings
from pathlib import Path

import numpy as np
import xarray as xr
from vtkmodules.vtkCommonDataModel import vtkDataObject
from vtkmodules.vtkIOLegacy import vtkDataSetReader
from vtkmodules.vtkIOXML import (
    vtkXMLImageDataReader,
    vtkXMLRectilinearGridReader,
    vtkXMLStructuredGridReader,
)
from xarray.backends import BackendEntrypoint

from pan3d.xarray.errors import DataCopyWarning

READERS = {
    ".vti": vtkXMLImageDataReader,
    ".vtr": vtkXMLRectilinearGridReader,
    ".vts": vtkXMLStructuredGridReader,
    ".vtk": vtkDataSetReader,
}


def read(file_path):
    reader = READERS[Path(file_path).suffix](file_name=file_path)
    # reader.SetFileName(file_path)
    return reader()


def rectilinear_grid_to_dataset(mesh):
    dims = list(mesh.dimensions)
    dims = dims[-1:] + dims[:-1]
    return xr.Dataset(
        {
            name: (["z", "x", "y"], mesh.point_data[name].ravel().reshape(dims))
            for name in mesh.point_data.keys()
        },
        coords={
            "x": (["x"], mesh.x_coordinates),
            "y": (["y"], mesh.y_coordinates),
            "z": (["z"], mesh.z_coordinates),
        },
    )


def image_data_to_dataset(mesh):
    origin = mesh.origin
    spacing = mesh.spacing
    extent = mesh.extent

    def gen_coords(i):
        return np.array(
            [
                origin[i] + (spacing[i] * v)
                for v in range(extent[i * 2], extent[i * 2 + 1] + 1)
            ],
            dtype=np.double,
        )

    dims = list(mesh.dimensions)
    return xr.Dataset(
        {
            name: (["z", "x", "y"], mesh.point_data[name].ravel().reshape(dims))
            for name in mesh.point_data.keys()
        },
        coords={
            "x": (["x"], gen_coords(0)),
            "y": (["y"], gen_coords(1)),
            "z": (["z"], gen_coords(2)),
        },
    )


def structured_grid_to_dataset(mesh):
    warnings.warn(
        DataCopyWarning(
            "StructuredGrid dataset engine duplicates data - VTK data not shared with xarray."
        ),
        stacklevel=2,
    )
    dims = [0, 0, 0]
    mesh.GetDimensions(dims)
    return xr.Dataset(
        {
            name: (
                ["xi", "yi", "zi"],
                mesh.point_data[name].ravel().reshape(dims),
            )
            for name in mesh.point_data.keys()
        },
        coords={
            "x": (["xi", "yi", "zi"], mesh.x_coordinates),
            "y": (["xi", "yi", "zi"], mesh.y_coordinates),
            "z": (["xi", "yi", "zi"], mesh.z_coordinates),
        },
    )


DATASET_TO_XARRAY = {
    "vtkRectilinearGrid": rectilinear_grid_to_dataset,
    "vtkImageData": image_data_to_dataset,
    "vtkStructuredGrid": structured_grid_to_dataset,
}


def dataset_to_xarray(dataset):
    if isinstance(dataset, vtkDataObject):
        for ds_type, fn in DATASET_TO_XARRAY.items():
            if dataset.IsA(ds_type):
                return fn(dataset)

    msg = f"pan3d is unable to generate an xarray DataSet from the {type(dataset)} VTK data type at this time."
    raise TypeError(msg)


class VTKBackendEntrypoint(BackendEntrypoint):
    def open_dataset(
        self,
        filename_or_obj,
        *,
        drop_variables=None,
    ):
        return dataset_to_xarray(read(filename_or_obj))

    open_dataset_parameters = [
        "filename_or_obj",
        "attrs",
    ]

    def guess_can_open(self, filename_or_obj):
        try:
            return Path(filename_or_obj).suffix in READERS
        except TypeError:
            return False



from vtkmodules.vtkCommonCore import vtkLookupTable

from pan3d.ui.css import base, vtk_view
from pan3d.utils.convert import to_image
from pan3d.utils.presets import PRESETS, set_preset
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class ScalarBar(v3.VTooltip):
    """
    Scalar bar for the XArray Explorers.
    """

    _next_id = 0

    @classmethod
    def next_id(cls):
        """Get the next unique ID for the scalar bar."""
        cls._next_id += 1
        return f"pan3d_scalarbar{cls._next_id}"

    def __init__(
        self, preset="Fast", color_min=0.0, color_max=1.0, ctx_name=None, **kwargs
    ):
        """Scalar bar for the XArray Explorers."""
        self._lut = vtkLookupTable()
        super().__init__(location="top", ctx_name=ctx_name)
        # Activate CSS
        self.server.enable_module(base)
        self.server.enable_module(vtk_view)

        ns = self.next_id()
        self.__preset_image = f"{ns}_preset"
        self.__color_min = f"{ns}_color_min"
        self.__color_max = f"{ns}_color_max"
        # Probe enables mouse events for scalar bar
        self.__probe_location = f"{ns}_probe_location"
        self.__probe_enabled = f"{ns}_probe_enabled"

        # Initialize state
        self.preset = preset
        self.set_color_range(color_min, color_max)
        self.state[self.__probe_location] = []
        self.state[self.__probe_enabled] = 0
        self.state.client_only(
            self.__probe_location,
            self.__probe_enabled,
        )

        with self:
            # Content
            with html.Template(v_slot_activator="{ props }"):
                with html.Div(
                    classes="scalarbar",
                    rounded="pill",
                    v_bind="props",
                    **kwargs,
                ):
                    html.Div(
                        f"{{{{{self.__color_min}.toFixed(6) }}}}",
                        classes="scalarbar-left",
                    )
                    html.Img(
                        src=(self.__preset_image, None),
                        style="height: 100%; width: 100%;",
                        classes="rounded-lg border-thin",
                        mousemove=f"{self.__probe_location} = [$event.x, $event.target.getBoundingClientRect()]",
                        mouseenter=f"{self.__probe_enabled} = 1",
                        mouseleave=f"{self.__probe_enabled} = 0",
                        __events=["mousemove", "mouseenter", "mouseleave"],
                    )
                    html.Div(
                        v_show=(self.__probe_enabled, False),
                        classes="scalar-cursor",
                        style=(
                            f"`left: ${{{self.__probe_location}?.[0] - {self.__probe_location}?.[1]?.left}}px`",
                        ),
                    )
                    html.Div(
                        f"{{{{ {self.__color_max}.toFixed(6) }}}}",
                        classes="scalarbar-right",
                    )
            html.Span(
                f"{{{{ (({self.__color_max} - {self.__color_min}) * ({self.__probe_location}?.[0] - {self.__probe_location}?.[1]?.left) / {self.__probe_location}?.[1]?.width + {self.__color_min}).toFixed(6) }}}}"
            )

    @property
    def preset(self):
        return self._preset

    @preset.setter
    def preset(self, value):
        if value not in PRESETS:
            err_msg = f"Preset '{value}' not found."
            raise ValueError(err_msg)
        self._preset = value
        set_preset(self._lut, value)
        with self.state:
            self.state[self.__preset_image] = to_image(self._lut)

    @property
    def preset_image_name(self):
        return self.__preset_image

    @property
    def color_min(self):
        return self.state[self.__color_min]

    @color_min.setter
    def color_min(self, value):
        with self.state:
            self.state[self.__color_min] = value

    @property
    def color_min_name(self):
        return self.__color_min

    @property
    def color_max(self):
        return self.state[self.__color_max]

    @color_max.setter
    def color_max(self, value):
        with self.state:
            self.state[self.__color_max] = value

    @property
    def color_max_name(self):
        return self.__color_max

    def set_color_range(self, color_min, color_max):
        """Set the color range for the scalar bar."""
        self.state[self.__color_min] = color_min
        self.state[self.__color_max] = color_max



import base64
import math

from vtkmodules.vtkCommonCore import vtkUnsignedCharArray
from vtkmodules.vtkCommonDataModel import vtkImageData
from vtkmodules.vtkIOImage import vtkPNGWriter


def to_float(v):
    v = float(v)
    if math.isnan(v) or v < 0.0001:
        v = 0.0001
    return v


def max_str_length(labels):
    size = 0
    for label in labels:
        s = len(label)
        size = max(s, size)

    return size


def update_camera(camera, props):
    for k, v in props.items():
        setattr(camera, k, v)


def to_image(lut, samples=255):
    colorArray = vtkUnsignedCharArray()
    colorArray.SetNumberOfComponents(3)
    colorArray.SetNumberOfTuples(samples)

    dataRange = lut.GetRange()
    delta = (dataRange[1] - dataRange[0]) / float(samples)

    # Add the color array to an image data
    imgData = vtkImageData()
    imgData.SetDimensions(samples, 1, 1)
    imgData.GetPointData().SetScalars(colorArray)

    # Loop over all presets
    rgb = [0, 0, 0]
    for i in range(samples):
        lut.GetColor(dataRange[0] + float(i) * delta, rgb)
        r = int(round(rgb[0] * 255))
        g = int(round(rgb[1] * 255))
        b = int(round(rgb[2] * 255))
        colorArray.SetTuple3(i, r, g, b)

    writer = vtkPNGWriter()
    writer.WriteToMemoryOn()
    writer.SetInputData(imgData)
    writer.SetCompressionLevel(6)
    writer.Write()

    writer.GetResult()

    base64_img = base64.standard_b64encode(writer.GetResult()).decode("utf-8")
    return f"data:image/png;base64,{base64_img}"



from trame.widgets import html
from trame.widgets import vuetify3 as vuetify


class CatalogSearch(vuetify.VDialog):
    def __init__(
        self,
        update_catalog_search_term_function,
        catalog_search_function,
        catalog_term_search_function,
        catalog="catalog",
        available_catalogs="available_catalogs",
        ui_search_catalogs="ui_search_catalogs",
        catalog_current_search="catalog_current_search",
        ui_catalog_term_search_loading="ui_catalog_term_search_loading",
        ui_catalog_search_message="ui_catalog_search_message",
    ):
        super().__init__(v_model=(ui_search_catalogs,), max_width=800)

        self.state.setdefault(ui_search_catalogs, False)
        self.state.setdefault(available_catalogs, [])
        self.state.setdefault(catalog_current_search, {})
        self.state.setdefault(ui_catalog_term_search_loading, False)
        self.state.setdefault(ui_catalog_search_message, "")

        with self:
            with vuetify.VCard():
                with vuetify.VCardText():
                    vuetify.VBtn(
                        flat=True,
                        icon="mdi-close",
                        style="float: right",
                        click=f"{ui_search_catalogs} = false",
                    )
                    vuetify.VCardTitle("Search a catalog")

                    vuetify.VSelect(
                        label="Choose a catalog",
                        v_show=(f"{available_catalogs}.length > 1",),
                        items=(available_catalogs, []),
                        v_model=(catalog,),
                        item_title="name",
                        density="compact",
                        return_object=True,
                        hide_details=True,
                    )

                    vuetify.VCardSubtitle(
                        "Select or enter search terms",
                        classes="mt-5",
                        style="display: inline-block",
                    )
                    with vuetify.VTooltip(
                        location="bottom",
                        text="For each search term, you may specify one or more acceptable values.",
                    ):
                        with vuetify.Template(v_slot_activator=("{ props }",)):
                            vuetify.VIcon(
                                v_bind="props",
                                icon="mdi-information-outline",
                                style="vertical-align: baseline",
                            )

                    with vuetify.VBtn(
                        size="small",
                        classes="mt-3",
                        style="float: right",
                        loading=(ui_catalog_term_search_loading,),
                        disabled=(ui_catalog_term_search_loading,),
                        click=catalog_term_search_function,
                    ):
                        html.Span("Search for term options")
                        vuetify.VTooltip(
                            activator="parent",
                            location="bottom",
                            text="""
                                Perform an unfiltered search on the catalog
                                to find all unique values for all searchable terms.
                                The results will appear as selectable options in
                                the value input dropdowns.
                                """,
                        )

                    with vuetify.VTable(v_if=(catalog,), style="max-height: 500px"):
                        with html.Tbody():
                            with html.Tr(
                                v_for=(f"term in {catalog}.search_terms",),
                            ):
                                html.Td("{{ term.key }}", style="width: 100px")
                                with html.Td():
                                    vuetify.VCombobox(
                                        label="Enter value",
                                        items=("term.options", []),
                                        multiple=True,
                                        chips=True,
                                        closable_chips=True,
                                        clearable=True,
                                        model_value=(
                                            f"{catalog_current_search}[term.key]",
                                            [],
                                        ),
                                        update_modelValue=(
                                            update_catalog_search_term_function,
                                            "[term.key, $event]",
                                        ),
                                    )

                    with vuetify.VBtn(
                        color="primary",
                        style="float: right",
                        loading=(ui_catalog_term_search_loading,),
                        disabled=(ui_catalog_term_search_loading,),
                        click=catalog_search_function,
                    ):
                        html.Span("Search")
                        vuetify.VTooltip(
                            activator="parent",
                            location="bottom",
                            text="""
                                Perform a filtered search with the union of all selected terms.
                                The results will be grouped and added to the list of available datasets.
                                """,
                        )

                    html.Span(
                        "{{ %s }}" % ui_catalog_search_message,
                        v_show=ui_catalog_search_message,
                    )



from pan3d.ui.css import base, vtk_view
from pan3d.utils.constants import VIEW_UPS
from trame.decorators import TrameApp, change
from trame.widgets import html
from trame.widgets import vtk as vtkw
from trame.widgets import vtklocal as wasm
from trame.widgets import vuetify3 as v3


@TrameApp()
class Pan3DView(html.Div):
    def __init__(
        self,
        render_window,
        import_pending="import_pending",
        axis_names="axis_names",
        local_rendering=None,
        widgets=None,
        disable_style_toggle=False,
        disable_roll=False,
        disable_axis_align=False,
        **kwargs,
    ):
        super().__init__(classes="pan3d-view", **kwargs)
        self.toolbar = None
        # Activate CSS
        self.server.enable_module(base)
        self.server.enable_module(vtk_view)

        # Initialize conditional widgets
        vtkw.initialize(self.server)
        wasm.initialize(self.server)

        self._import_pending = import_pending
        self.render_window = render_window
        self.renderer = render_window.GetRenderers().GetFirstRenderer()
        self.camera = self.renderer.active_camera

        # Expose view_reset_clipping_range
        self.ctrl.view_reset_clipping_range = self.renderer.ResetCameraClippingRange

        # Reserved state with default
        self.state.setdefault("view_3d", True)
        self.state.setdefault(import_pending, False)

        with self:
            # 3D view
            if local_rendering is not None:
                if local_rendering == "wasm":
                    with wasm.LocalView(
                        self.render_window,
                        throttle_rate=10,
                        listeners=("wasm_listeners", {}),
                    ) as view:
                        self.ctrl.view_update_force = view.update
                        self.ctrl.view_update = view.update_throttle
                        self.ctrl.view_reset_camera = view.reset_camera
                        camera_id = view.get_wasm_id(self.camera)
                        self.state.setdefault("wasm_camera", None)
                        self.state.wasm_listeners = {
                            camera_id: {
                                "ModifiedEvent": {
                                    "wasm_camera": {
                                        "position": [camera_id, "Position"],
                                        "view_up": [camera_id, "ViewUp"],
                                        "focal_point": [camera_id, "FocalPoint"],
                                    }
                                }
                            }
                        }
                        for w in widgets or []:
                            view.register_widget(w)
                else:
                    with vtkw.VtkLocalView(self.render_window) as view:
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera
                        view.set_widgets(widgets or [])
            else:
                with vtkw.VtkRemoteView(
                    self.render_window, interactive_ratio=1
                ) as view:
                    self.ctrl.view_update = view.update
                    self.ctrl.view_reset_camera = view.reset_camera

            # Scroll locking overlay
            html.Div(v_show=("view_locked", False), classes="view-lock")

            # 3D toolbox
            with v3.VCard(classes="view-toolbar pa-1", rounded="lg") as toolbar:
                self.toolbar = toolbar
                with v3.VTooltip(text="Lock view interaction"):
                    with html.Template(v_slot_activator="{ props }"):
                        v3.VBtn(
                            v_bind="props",
                            flat=True,
                            density="compact",
                            icon=(
                                "view_locked ? 'mdi-lock-outline' : 'mdi-lock-off-outline'",
                            ),
                            click="view_locked = !view_locked",
                        )
                v3.VDivider(classes="my-1")
                with v3.VTooltip(text="Reset camera"):
                    with html.Template(v_slot_activator="{ props }"):
                        v3.VBtn(
                            v_bind="props",
                            flat=True,
                            density="compact",
                            icon="mdi-crop-free",
                            click=self.ctrl.view_reset_camera,
                        )

                if not (disable_style_toggle and disable_roll and disable_axis_align):
                    v3.VDivider(classes="my-1")

                if not disable_style_toggle:
                    with v3.VTooltip(text="Toggle between 3D/2D interaction"):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon=(
                                    "view_3d ? 'mdi-rotate-orbit' : 'mdi-cursor-move'",
                                ),
                                click="view_3d = !view_3d",
                            )
                    v3.VDivider(classes="my-1")
                if not disable_roll:
                    with v3.VTooltip(text="Rotate left 90"):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-rotate-left",
                                click=(self.rotate_camera, "[-1]"),
                            )
                    with v3.VTooltip(text="Rotate right 90"):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-rotate-right",
                                click=(self.rotate_camera, "[+1]"),
                            )
                    v3.VDivider(classes="my-1")

                if not disable_axis_align:
                    with v3.VTooltip(
                        text=(f"`Look toward ${{ {axis_names}[0] || 'X' }}`",)
                    ):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-axis-x-arrow",
                                click=(self.reset_camera_to_axis, "[[1,0,0]]"),
                            )
                    with v3.VTooltip(
                        text=(f"`Look toward ${{ {axis_names}[1] || 'Y'}}`",)
                    ):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-axis-y-arrow",
                                click=(self.reset_camera_to_axis, "[[0,1,0]]"),
                            )

                    with v3.VTooltip(
                        text=(f"`Look toward ${{ {axis_names}[2] || 'Z'}}`",)
                    ):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-axis-z-arrow",
                                click=(self.reset_camera_to_axis, "[[0,0,1]]"),
                            )
                    v3.VDivider(classes="my-1")
                    with v3.VTooltip(text="Look toward at an angle"):
                        with html.Template(v_slot_activator="{ props }"):
                            v3.VBtn(
                                v_bind="props",
                                flat=True,
                                density="compact",
                                icon="mdi-axis-arrow",
                                click=(self.reset_camera_to_axis, "[[1,1,1]]"),
                            )

    def reset_camera_to_axis(self, axis):
        camera = self.renderer.active_camera
        camera.focal_point = (0, 0, 0)
        camera.position = axis
        camera.view_up = VIEW_UPS.get(tuple(axis))
        self.renderer.ResetCamera()
        self.ctrl.view_update(push_camera=True)

    def rotate_camera(self, direction):
        camera = self.renderer.active_camera
        a = [*camera.view_up]
        b = [camera.focal_point[i] - camera.position[i] for i in range(3)]
        view_up = [
            direction * (a[1] * b[2] - a[2] * b[1]),
            direction * (a[2] * b[0] - a[0] * b[2]),
            direction * (a[0] * b[1] - a[1] * b[0]),
        ]
        camera.view_up = view_up

        self.ctrl.view_update(push_camera=True)

    @change("wasm_camera")
    def _on_camera(self, wasm_camera, **_):
        if wasm_camera is None:
            return

        for k, v in wasm_camera.items():
            setattr(self.camera, k, v)

    @change("view_3d")
    def _on_view_type_change(self, view_3d, **_):
        # FIXME properly swap interactor style
        if view_3d:
            # self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()
            self.renderer.GetActiveCamera().SetParallelProjection(0)
        else:
            # self.interactor.GetInteractorStyle().SetCu()
            self.renderer.GetActiveCamera().SetParallelProjection(1)

        if not self.state[self._import_pending]:
            self.ctrl.view_reset_camera()



from xarray.tutorial import open_dataset

ALL_ENTRIES = [
    {
        "name": "air_temperature",
        "description": "NCEP reanalysis subset",
    },
    {
        "name": "air_temperature_gradient",
        "description": "NCEP reanalysis subset with approximate x,y gradients",
    },
    {
        "name": "basin_mask",
        "description": "Dataset with ocean basins marked using integers",
    },
    # -------------------------------------------------------------------------
    # {
    #     "name": "ASE_ice_velocity",
    #     "description": "MEaSUREs InSAR-Based Ice Velocity of the Amundsen Sea Embayment, Antarctica, Version 1",
    # },
    # -------------------------------------------------------------------------
    # {
    #     "name": "rasm",
    #     "description": "Output of the Regional Arctic System Model (RASM)",
    # },
    # -------------------------------------------------------------------------
    # {
    #     "name": "ROMS_example",
    #     "description": "Regional Ocean Model System (ROMS) output",
    # },
    # -------------------------------------------------------------------------
    # {
    #     "name": "tiny",
    #     "description": "small synthetic dataset with a 1D data variable",
    # },
    # -------------------------------------------------------------------------
    # needs pandas[xarray]
    # {
    #     "name": "era5-2mt-2019-03-uk.grib",
    #     "description": "ERA5 temperature data over the UK",
    # },
    # -------------------------------------------------------------------------
    {
        "name": "eraint_uvz",
        "description": "data from ERA-Interim reanalysis, monthly averages of upper level data",
    },
    {
        "name": "ersstv5",
        "description": "NOAA's Extended Reconstructed Sea Surface Temperature monthly averages",
    },
]


def get_catalog():
    return {
        "name": "Xarray Tutorial",
        "id": "xarray",
        "search_terms": [{"key": "id", "options": []}],
    }


def get_search_options():
    search_options = {
        "name": [],
    }
    for entry_info in ALL_ENTRIES:
        for search_option, search_value in search_options.items():
            entry_value = entry_info.get(search_option)
            if entry_value:
                if isinstance(entry_value, str):
                    entry_value = [entry_value]
                search_options[search_option] = list({*search_value, *entry_value})

    return search_options


def to_result(entry):
    return {**entry, "value": {"source": "xarray", "id": entry["name"]}}


def search(**kwargs):
    results = [to_result(entry) for entry in ALL_ENTRIES]
    group_name = "Xarray Tutorial"
    message = (
        f'Found {len(ALL_ENTRIES)} dataset ids. Results added to group "{group_name}".'
    )
    return (results, group_name, message)


def load_dataset(id):
    return open_dataset(id)



import json
import traceback
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonDataModel import vtkRectilinearGrid
from vtkmodules.vtkFiltersCore import vtkArrayCalculator

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------


def get_time_labels(times):
    return [pd.to_datetime(time).strftime("%Y-%m-%d %H:%M:%S") for time in times]


def is_time_type(dtype):
    if np.issubdtype(dtype, np.datetime64):
        return True

    return np.issubdtype(dtype, np.dtype("O"))


def slice_array(array_name, dataset, slice_info):
    if array_name is None:
        return np.zeros(1, dtype=np.float32)
    array = dataset[array_name].to_numpy()
    if slice_info is None:
        return array
    if isinstance(slice_info, int):
        return array[slice_info]
    return array[slice(*slice_info)]


def to_isel(slices_info, *array_names):
    slices = {}
    for name in array_names:
        if name is None:
            continue

        info = slices_info.get(name)
        if info is None:
            continue
        if isinstance(info, int):
            slices[name] = info
        else:
            slices[name] = slice(*info)

    return slices if slices else None


# -----------------------------------------------------------------------------
# VTK Algorithms
# -----------------------------------------------------------------------------


class vtkXArrayRectilinearSource(VTKPythonAlgorithmBase):
    """vtkRectilinearGridAlgoritm for converting XArray as input"""

    def __init__(
        self,
        input: Optional[xr.Dataset] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
        z: Optional[str] = None,
        t: Optional[str] = None,
        arrays: Optional[list[str]] = None,
        order: str = "C",
    ):
        """
        Create vtkXArrayRectilinearSource

        Parameters:
            input (xr.Dataset): Provide an XArray to use as input. The load() method will replace it.
            x (str): Name of the dimension to use for X. The dimension needs to work with the selected arrays.
            y (str): Name of the dimension to use for Y. The dimension needs to work with the selected arrays.
            z (str): Name of the dimension to use for Z. The dimension needs to work with the selected arrays.
            t (str): Name of the dimension to use for time. The dimension needs to work with the selected arrays.
            arrays (list[str]): List of field to load onto the generated VTK mesh.
            order (str): C or F for the convention order (C or Fortran). (default: C)
        """
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkRectilinearGrid",
        )
        # Data source
        self._input = input
        self._xarray_mesh = None
        self._pipeline = None
        self._computed = {}
        self._data_origin = None

        # Array name selectors
        self._x = x
        self._y = y
        self._z = z
        self._t = t

        # Data sub-selection
        self._array_names = set(arrays or [])
        self._t_index = 0
        self._slices = None

        # Data order
        self._order = order

        # Auto apply coords if none provided
        if all(v is None for v in (x, y, z, t)):
            self.apply_coords()

        if len(self._array_names) == 0:
            self.arrays = self.available_arrays

    def __str__(self):
        return f"""
x: {self.x}
y: {self.y}
z: {self.z}
t: {self.t} ({self.t_index + 1}/{self.t_size})
loaded: {self.arrays}
all:    {self.available_arrays}
slices: {json.dumps(self.slices, indent=2)}
computed: {json.dumps(self.computed, indent=2)}
order: {self._order}
"""

    # -------------------------------------------------------------------------
    # Data input
    # -------------------------------------------------------------------------

    @property
    def input(self):
        """return current input XArray"""
        return self._input

    @input.setter
    def input(self, xarray_dataset: xr.Dataset):
        """update input with a new XArray"""
        self._input = xarray_dataset
        self._xarray_mesh = None
        self.Modified()

    # -------------------------------------------------------------------------
    # Array selectors
    # -------------------------------------------------------------------------

    @property
    def x(self):
        """return the name that is currently mapped to the X axis"""
        return self._x

    @x.setter
    def x(self, x_array_name: str):
        """update the coordinate name that is mapped to the X axis"""
        if x_array_name is None:
            if self._x is not None:
                self._x = None
                self._xarray_mesh = None
                self.Modified()
            return

        coords = self.available_coords
        if x_array_name not in coords:
            msg = f"x={x_array_name} is not a coordinate array [{', '.join(coords)}]"
            raise ValueError(msg)
        if self._x != x_array_name:
            self._x = x_array_name
            self._xarray_mesh = None
            self.Modified()

    @property
    def x_size(self):
        """return the size of the coordinate used for the X axis"""
        if self._x is None:
            return 0
        return int(self._input[self._x].size)

    @property
    def y(self):
        """return the name that is currently mapped to the Y axis"""
        return self._y

    @y.setter
    def y(self, y_array_name: str):
        """update the coordinate name that is mapped to the Y axis"""
        if y_array_name is None:
            if self._y is not None:
                self._y = None
                self._xarray_mesh = None
                self.Modified()
            return

        coords = self.available_coords
        if y_array_name not in coords:
            msg = f"y={y_array_name} is not a coordinate array [{', '.join(coords)}]"
            raise ValueError(msg)
        if self._y != y_array_name:
            self._y = y_array_name
            self._xarray_mesh = None
            self.Modified()

    @property
    def y_size(self):
        """return the size of the coordinate used for the Y axis"""
        if self._y is None:
            return 0
        return int(self._input[self._y].size)

    @property
    def z(self):
        """return the name that is currently mapped to the Z axis"""
        return self._z

    @z.setter
    def z(self, z_array_name: str):
        """update the coordinate name that is mapped to the Z axis"""
        if z_array_name is None:
            if self._z is not None:
                self._z = None
                self._xarray_mesh = None
                self.Modified()
            return

        coords = self.available_coords
        if z_array_name not in coords:
            msg = f"z={z_array_name} is not a coordinate array [{', '.join(coords)}]"
            raise ValueError(msg)
        if self._z != z_array_name:
            self._z = z_array_name
            self._xarray_mesh = None
            self.Modified()

    @property
    def z_size(self):
        """return the size of the coordinate used for the Z axis"""
        if self._z is None:
            return 0
        return int(self._input[self._z].size)

    @property
    def t(self):
        """return the name that is currently mapped to the time axis"""
        return self._t

    @t.setter
    def t(self, t_array_name: str):
        """update the coordinate name that is mapped to the time axis"""
        if t_array_name is None:
            if self._t is not None:
                self._t = None
                self._xarray_mesh = None
                self.Modified()
            return

        coords = self.available_coords
        if t_array_name not in coords:
            msg = f"t={t_array_name} is not a coordinate array [{', '.join(coords)}]"
            raise ValueError(msg)
        if self._t != t_array_name:
            self._t = t_array_name
            self._xarray_mesh = None
            self.Modified()

    @property
    def slice_extents(self):
        """return a dictionary for the X, Y, Z dimensions with the corresponding extent [0, size-1]"""
        return {
            coord_name: [0, self.input[coord_name].size - 1]
            for coord_name in [self.x, self.y, self.z]
            if coord_name is not None
        }

    def apply_coords(self):
        """Use array dims to automatically map coordinates (x,y,z,t)"""
        if self.input is None:
            return

        array_name = self.available_arrays[0]
        coords = self._input[array_name].dims

        # reset coords arrays
        self.x = None
        self.y = None
        self.z = None
        self.t = None

        # assign mapping
        axes = ["t", "z", "y", "x"]
        if len(coords) == 4:
            for key, value in zip(axes, coords):
                setattr(self, key, value)
        elif len(coords) == 2:
            axes.remove("t")
            axes.remove("z")
            for key, value in zip(axes, coords):
                setattr(self, key, value)
        elif len(coords) == 3:
            # Is it 2D dataset with time or 3D dataset ?
            outer_dtype = self._input[array_name][coords[0]].dtype
            if is_time_type(outer_dtype):
                axes.remove("z")
                for key, value in zip(axes, coords):
                    setattr(self, key, value)
            else:
                axes.remove("t")
                for key, value in zip(axes, coords):
                    setattr(self, key, value)

    @property
    def available_coords(self):
        """List available coordinates arrays that have are 1D"""
        if self._input is None:
            return []

        return [k for k, v in self._input.coords.items() if len(v.shape) == 1]

    # -------------------------------------------------------------------------
    # Data sub-selection
    # -------------------------------------------------------------------------

    @property
    def t_index(self):
        """return the current selected time index"""
        return self._t_index

    @t_index.setter
    def t_index(self, t_index: int):
        """update the current selected time index"""
        if t_index != self._t_index:
            self._t_index = t_index
            self._xarray_mesh = None
            self.Modified()

    @property
    def t_size(self):
        """return the size of the coordinate used for the time"""
        if self._t is None:
            return 0
        return int(self._input[self._t].size)

    @property
    def t_labels(self):
        """return a list of string that match the various time values available"""
        if self._t is None:
            return []

        t_array = self._input[self._t]
        t_type = t_array.dtype
        if np.issubdtype(t_type, np.datetime64):
            return get_time_labels(t_array.to_numpy())
        return [str(t) for t in t_array.to_numpy()]

    @property
    def arrays(self):
        """return the list of arrays that are currently selected to be added to the generated VTK mesh"""
        return list(self._array_names)

    @arrays.setter
    def arrays(self, array_names: list[str]):
        """update the list of arrays to load on the generated VTK mesh"""
        new_names = set(array_names or [])
        if new_names != self._array_names:
            self._array_names = new_names
            self._xarray_mesh = None
            self.Modified()

    @property
    def available_arrays(self):
        """List all available data fields for the `arrays` option"""
        if self._input is None:
            return []

        filtered_arrays = []
        max_dim = 0
        coords = set(self.available_coords)
        for name in set(self._input.data_vars.keys()) - set(self._input.coords.keys()):
            if name.endswith(("_bnds", "_bounds")):
                continue

            dims = set(self._input[name].dims)
            max_dim = max(max_dim, len(dims))
            if dims.issubset(coords):
                filtered_arrays.append(name)

        return [n for n in filtered_arrays if len(self._input[n].shape) == max_dim]

    @property
    def slices(self):
        """return the current slicing information which include axes crop/cut and time selection"""
        result = dict(self._slices or {})
        if self.t is not None:
            result[self.t] = self.t_index
        return result

    @slices.setter
    def slices(self, v):
        """update the slicing of the data along axes"""
        if v != self._slices:
            self._slices = v
            self._xarray_mesh = None
            self.Modified()

    # -------------------------------------------------------------------------
    # properties
    # -------------------------------------------------------------------------

    @property
    def order(self):
        """return the order used to decode numpy arrays"""
        return self._order

    @order.setter
    def order(self, order: str):
        """update the order to use for decoding numpy arrays"""
        self._order = order
        self._xarray_mesh = None
        self.Modified()

    # -------------------------------------------------------------------------
    # add-on logic
    # -------------------------------------------------------------------------

    @property
    def computed(self):
        """return the current description of the computed/derived fields on the VTK mesh"""
        return self._computed

    @computed.setter
    def computed(self, v):
        """
        update the computed/derived fields to add on the VTK mesh

        The layout of the dictionary provided should be as follow:
          - key: name of the field to be added
          - value: formula to apply for the given field name. The syntax is captured in the document (https://docs.paraview.org/en/latest/UsersGuide/filteringData.html#calculator)

        Then additional keys need to be provided to describe your formula dependencies:
        `_use_scalars` and `_use_vectors` which should be a list of string matching the name of the fields you are using in your expression.


        Please find below an example:

        ```
            {
                "_use_scalars": ["u", "v"],       # (u,v) needed for "vec" and "m2"
                "vec": "(u * iHat) + (v * jHat)", # 2D vector
                "m2": "u*u + v*v",
            }
        ```
        """
        if self._computed != v:
            self._computed = v or {}
            self._pipeline = None
            scalar_arrays = self._computed.get("_use_scalars", [])
            vector_arrays = self._computed.get("_use_vectors", [])

            for output_name, func in self._computed.items():
                if output_name[0] == "_":
                    continue
                filter = vtkArrayCalculator(
                    result_array_name=output_name,
                    function=func,
                )

                # register array dependencies
                for scalar_array in scalar_arrays:
                    filter.AddScalarArrayName(scalar_array)
                for vector_array in vector_arrays:
                    filter.AddVectorArrayName(vector_array)

                if self._pipeline is None:
                    self._pipeline = filter
                else:
                    self._pipeline = self._pipeline >> filter

            self.Modified()

    def load(self, data_info):
        """
        create a new XArray input with the `data_origin` and `dataset_config` information.

        Here is an example of the layout of the parameter

        ```
        {
          "data_origin": {
            "source": "url", # one of [file, url, xarray, pangeo, esgf]
            "id": "https://ncsa.osn.xsede.org/Pangeo/pangeo-forge/noaa-coastwatch-geopolar-sst-feedstock/noaa-coastwatch-geopolar-sst.zarr",
            "order": "C"     # (optional) order to use in numpy
          },
          "dataset_config": {
            "x": "lon",      # (optional) coord name for X
            "y": "lat",      # (optional) coord name for Y
            "z": null,       # (optional) coord name for Z
            "t": "time",     # (optional) coord name for time
            "slices": {      # (optional) array slicing
              "lon": [
                1000,
                6000,
                20
              ],
              "lat": [
                500,
                3000,
                20
              ],
              "time": 5
            },
            "t_index": 5,    # (optional) selected time index
            "arrays": [      # (optional) names of arrays to load onto VTK mesh.
              "analysed_sst" #            If missing no array will be loaded
            ]                #            onto the mesh.
          }
        }
        ```
        """
        if "data_origin" not in data_info:
            raise ValueError("Only state with data_origin can be loaded")

        from pan3d import catalogs

        self._data_origin = data_info["data_origin"]
        self.input = catalogs.load_dataset(
            self._data_origin["source"], self._data_origin["id"]
        )

        self.order = self._data_origin.get("order", "C")

        dataset_config = data_info.get("dataset_config")
        if dataset_config is None:
            self.apply_coords()
            self.arrays = self.available_arrays
        else:
            self.x = dataset_config.get("x")
            self.y = dataset_config.get("y")
            self.z = dataset_config.get("z")
            self.t = dataset_config.get("t")
            self.slices = dataset_config.get("slices")
            self.t_index = dataset_config.get("t_index", 0)
            self.apply_coords()
            self.arrays = dataset_config.get("arrays", self.available_arrays)

    @property
    def state(self):
        """return current state that can be reused in a load() later on"""
        if self._data_origin is None:
            raise RuntimeError(
                "No state available without data origin. Need to use the load method to set the data origin."
            )

        return {
            "data_origin": self._data_origin,
            "dataset_config": {
                k: getattr(self, k)
                for k in ["x", "y", "z", "t", "slices", "t_index", "arrays"]
            },
        }

    # -------------------------------------------------------------------------
    # Algorithm
    # -------------------------------------------------------------------------

    def RequestData(self, request, inInfo, outInfo):
        """implementation of the vtk algorithm for generating the VTK mesh"""
        # Use open data_array handle to fetch data at
        # desired Level of Detail
        try:
            pdo = self.GetOutputData(outInfo, 0)

            # Generate mesh
            if self._xarray_mesh is None:
                # grid
                mesh = vtkRectilinearGrid()
                mesh.x_coordinates = slice_array(
                    self._x, self._input, self.slices.get(self._x)
                )
                mesh.y_coordinates = slice_array(
                    self._y, self._input, self.slices.get(self._y)
                )
                mesh.z_coordinates = slice_array(
                    self._z, self._input, self.slices.get(self._z)
                )
                mesh.dimensions = [
                    mesh.x_coordinates.size,
                    mesh.y_coordinates.size,
                    mesh.z_coordinates.size,
                ]
                # fields
                indexing = to_isel(self.slices, self.x, self.y, self.z, self.t)
                for field_name in self._array_names:
                    da = self._input[field_name]
                    if indexing is not None:
                        da = da.isel(indexing)
                    mesh.point_data[field_name] = da.to_numpy().ravel(order=self._order)

                self._xarray_mesh = mesh

            # Compute derived quantity
            if self._pipeline is not None:
                pdo.ShallowCopy(self._pipeline(self._xarray_mesh))
            else:
                pdo.ShallowCopy(self._xarray_mesh)

        except Exception as e:
            traceback.print_exc()
            raise e
        return 1



from datetime import datetime

import intake

CATALOG_URL = "https://raw.githubusercontent.com/pangeo-data/pangeo-datastore/master/intake-catalogs/master.yaml"


def get_catalog():
    return {
        "name": "Pangeo Forge",
        "id": "pangeo",
        "search_terms": [{"key": "id", "options": []}],
    }


def get_entry_info(entry):
    entry_data = entry.describe()
    metadata = entry_data.get("metadata", {})
    args = entry_data.get("args", {})

    name = entry_data.get("name")
    description = entry_data.get("description")
    container = entry_data.get("container")
    driver_list = entry_data.get("driver", [])
    tags_list = metadata.get("tags", [])
    requester_pays = args.get("storage_options", {}).get("requester_pays", False)
    requester_pays = str(requester_pays).lower()

    return {
        "name": name,
        "description": description,
        "container": container,
        "driver": driver_list,
        "tags": tags_list,
        "requester_pays": requester_pays,
    }


def get_all_entries():
    all_entries = []
    catalog = intake.open_catalog(CATALOG_URL)
    for subcatalog_name in catalog:
        subcatalog = catalog[subcatalog_name]
        for entry_name, entry_data in subcatalog._entries.items():
            if entry_data.container == "catalog":
                for subentry_name in entry_data:
                    subentry = entry_data[subentry_name]
                    subentry_info = get_entry_info(subentry)
                    subentry_info["catalog"] = f"{subcatalog_name}/{entry_name}"
                    all_entries.append(subentry_info)
            else:
                entry_info = get_entry_info(entry_data)
                entry_info["catalog"] = subcatalog_name
                all_entries.append(entry_info)
    return all_entries


def entry_filter_match(entry, filters):
    for key, selected_values in filters.items():
        # name is the unique identifier
        filter_key = "name" if key == "id" else key
        entry_value = entry.get(filter_key)
        if entry_value is None:
            return False
        if isinstance(entry_value, str):
            entry_value = [entry_value]
        if len(selected_values) and not any(v in selected_values for v in entry_value):
            return False

    return True


def get_search_options():
    all_entries = get_all_entries()
    search_options = {
        "name": [],
        "container": [],
        "driver": [],
        "tags": [],
        "catalog": [],
        "requester_pays": ["true", "false"],
    }
    for entry_info in all_entries:
        for search_option, search_value in search_options.items():
            entry_value = entry_info.get(search_option)
            if entry_value:
                if isinstance(entry_value, str):
                    entry_value = [entry_value]
                search_options[search_option] = list({*search_value, *entry_value})

    return search_options


def search(**kwargs):
    group_name = "/".join([f"{k}:{','.join(v)}" for k, v in kwargs.items()])
    if not group_name:
        group_name = "All Pangeo Datasets"
    start = datetime.now()
    all_entries = get_all_entries()
    results = [
        {
            "id": entry["name"],
            "subtitle": entry["description"],
            "value": {"source": "pangeo", "id": entry["name"]},
        }
        for entry in all_entries
        if entry_filter_match(entry, kwargs)
    ]

    delta = datetime.now() - start
    message = f'Found {len(results)} dataset ids \
        in {delta.total_seconds()} seconds.\
        Results added to group "{group_name}".'
    return (results, group_name, message)


def informative_error(e, **kwargs):
    if isinstance(
        e, ValueError
    ) and "User project specified in the request is invalid" in str(e):
        return "Dataset has requester pays policy and no billable Google Cloud Project was specified. See https://github.com/pangeo-data/pangeo-datastore?tab=readme-ov-file#accessing-requester-pays-data for more info."
    if "GS_SECRET_ACCESS_KEY+GS_ACCESS_KEY_ID" in str(e):
        return "Dataset requires Google Cloud authentication. See https://cloud.google.com/sdk/gcloud/reference/auth/login for more info."
    if "Unable to locate credentials" in str(e):
        return "Dataset requires AWS authentication. See https://docs.aws.amazon.com/signin/latest/userguide/command-line-sign-in.html for more info."

    return f"Failed to load dataset {kwargs.get('dataset')} - {e!s}"


def load_dataset(id):
    catalog = intake.open_catalog(CATALOG_URL)
    for subcatalog_name in catalog:
        subcatalog = catalog[subcatalog_name]
        for entry_name, entry_data in subcatalog._entries.items():
            try:
                entry = subcatalog[entry_name]
                if entry.container == "catalog":
                    for subentry_name in entry:
                        if subentry_name == id:
                            subentry = entry[subentry_name]
                            return subentry.to_dask()
                elif entry_name == id:
                    return entry.to_dask()
            except Exception as e:
                raise ValueError(
                    informative_error(e, dataset=entry_data.describe())
                ) from e
    return None



import math

from pan3d.utils.common import RenderingSettingsBasic
from pan3d.utils.convert import max_str_length
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class SliceRenderingSettings(RenderingSettingsBasic):
    def __init__(self, source, update_rendering, **kwargs):
        super().__init__(source, update_rendering, **kwargs)

        self.source = source

        style = {"density": "compact", "hide_details": True}
        with self.content:
            # -- Slice section
            v3.VDivider()
            with v3.VRow(classes="mx-2 my-0 align-center"):
                html.Span("Slice", classes="text-h6 font-weight-medium")
                v3.VSpacer()
                html.Span(
                    "{{parseFloat(cut_x).toFixed(2)}}",
                    v_show="slice_axis === axis_names[0]",
                    classes="text-subtitle-1",
                )
                html.Span(
                    "{{parseFloat(cut_y).toFixed(2)}}",
                    v_show="slice_axis === axis_names[1]",
                    classes="text-subtitle-1",
                )
                html.Span(
                    "{{parseFloat(cut_z).toFixed(2)}}",
                    v_show="slice_axis === axis_names[2]",
                    classes="text-subtitle-1",
                )
            with v3.VRow(classes="mx-2 my-0"):
                v3.VSelect(
                    v_model=("slice_axis",),
                    items=("axis_names",),
                    **style,
                )
            with v3.VRow(classes="mx-2 my-0"):
                v3.VSlider(
                    v_show="slice_axis === axis_names[0]",
                    v_model=("cut_x",),
                    min=("bounds[0]",),
                    max=("bounds[1]",),
                    **style,
                )
                v3.VSlider(
                    v_show="slice_axis === axis_names[1]",
                    v_model=("cut_y",),
                    min=("bounds[2]",),
                    max=("bounds[3]",),
                    **style,
                )
                v3.VSlider(
                    v_show="slice_axis === axis_names[2]",
                    v_model=("cut_z",),
                    min=("bounds[4]",),
                    max=("bounds[5]",),
                    **style,
                )

            with v3.VRow(classes="mx-2 my-0"):
                with v3.VCol():
                    html.Div(
                        "{{parseFloat(bounds[axis_names.indexOf(slice_axis)*2]).toFixed(2)}}",
                        classes="font-weight-medium",
                    )
                with v3.VCol(classes="text-right"):
                    html.Div(
                        "{{parseFloat(bounds[axis_names.indexOf(slice_axis)*2 + 1]).toFixed(2)}}",
                        classes="font-weight-medium",
                    )

            v3.VDivider()
            # Actor scaling
            with v3.VTooltip(text="Representation scaling"):
                with html.Template(v_slot_activator="{ props }"):
                    with v3.VRow(
                        v_bind="props",
                        no_gutter=True,
                        classes="align-center my-0 mx-0 border-b-thin",
                    ):
                        v3.VIcon(
                            "mdi-ruler-square",
                            classes="ml-2 text-medium-emphasis",
                        )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[0]"):
                            v3.VTextField(
                                v_model=("scale_x", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[1]"):
                            v3.VTextField(
                                v_model=("scale_y", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[2]"):
                            v3.VTextField(
                                v_model=("scale_z", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )
            v3.VDivider()
            # Time slider
            with v3.VTooltip(
                v_if="slice_t_max > 0",
                text=("`time: ${slice_t + 1} / ${slice_t_max+1}`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            prepend_icon="mdi-clock-outline",
                            v_model=("slice_t", 0),
                            min=0,
                            max=("slice_t_max", 0),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
            v3.VDivider()
            v3.VBtn(
                "Update 3D view",
                block=True,
                classes="text-none",
                flat=True,
                density="compact",
                rounded=0,
                disabled=("data_arrays.length === 0",),
                color=("dirty_data && data_arrays.length ? 'primary': undefined",),
                click=(update_rendering, "[true]"),
            )

    def update_from_source(self, source=None):
        if source is None:
            return

        ds = source()
        bounds = ds.bounds
        origin = [
            0.5 * (bounds[0] + bounds[1]),
            0.5 * (bounds[2] + bounds[3]),
            0.5 * (bounds[4] + bounds[5]),
        ]
        with self.state as state:
            state.data_arrays_available = source.available_arrays
            state.data_arrays = source.arrays

            state.color_by = None
            state.axis_names = [
                x for x in [source.x, source.y, source.z] if x is not None
            ]
            state.slice_extents = source.slice_extents

            # Update time
            state.slice_t = source.t_index
            state.slice_t_max = source.t_size - 1
            state.t_labels = source.t_labels
            state.max_time_width = math.ceil(0.58 * max_str_length(state.t_labels))

            if state.slice_t_max > 0:
                state.max_time_index_width = math.ceil(
                    0.6 + (math.log10(state.slice_t_max + 1) + 1) * 2 * 0.58
                )

            # Update state from dataset
            state.bounds = ds.bounds
            state.cut_x = origin[0]
            state.cut_y = origin[1]
            state.cut_z = origin[2]
            state.slice_axis = source.z if source.z is not None else source.y
            state.slice_axes = state.axis_names



# https://intake-esgf.readthedocs.io/en/latest/quickstart.html
from datetime import datetime

from intake_esgf import ESGFCatalog
from intake_esgf.exceptions import NoSearchResults


def get_catalog():
    return {
        "name": "ESGF",
        "id": "esgf",
        "search_terms": [{"key": "id", "options": []}],
    }


def get_search_options():
    catalog = ESGFCatalog()
    # perform unfiltered search and get unique values for each column
    results = catalog.search()
    search_options = {}
    for col in results.df.columns:
        search_options[col] = list(results.df[col].explode().unique())
    return search_options


def search(**kwargs):
    group_name = "/".join([f"{k}:{','.join(v)}" for k, v in kwargs.items()])
    if not group_name:
        group_name = "All ESGF Datasets"
    start = datetime.now()
    catalog = ESGFCatalog()
    results = []
    try:
        search = catalog.search(**kwargs)
        results = [
            {
                "id": id,
                "subtitle": id,
                "value": {"source": "esgf", "id": id},
            }
            for id in search.df.id.explode().unique()
        ]
    except NoSearchResults:
        pass

    delta = datetime.now() - start
    message = f'Found {len(results)} dataset ids \
        in {delta.total_seconds()} seconds.\
        Results added to group "{group_name}".'
    return (results, group_name, message)


def load_dataset(id):
    catalog = ESGFCatalog()
    search = catalog.search(id=id)
    loaded = search.to_dataset_dict()
    keys = list(loaded.keys())
    if len(keys) > 0:
        return loaded[keys[0]]
    msg = f"No dataset found for {id}."
    raise ValueError(msg)



from pathlib import Path

import xarray


def get_catalog():
    return {
        "name": "Local datasets",
        "id": "file",
        "search_terms": [{"key": "id", "options": []}],
    }


def get_search_options():
    return {"file": []}


def search(**kwargs):
    results = []
    group_name = "Local datasets"
    message = "Sorry but we don't have any search mechanism for that data source"
    return (results, group_name, message)


def load_dataset(id):
    if not Path(id).exists():
        msg = f"Could not find dataset at {id}"
        raise ValueError(msg)

    engine = None
    if ".zarr" in id:
        engine = "zarr"
    if ".nc" in id:
        engine = "netcdf4"
    return xarray.open_dataset(id, engine=engine, chunks={})



from trame.app import get_server
from trame.widgets import html
from trame.widgets import vuetify3 as vuetify

server = get_server()


class FileSelect(vuetify.VCard):
    def __init__(
        self,
        import_function,
        export_function,
        ui_action_name="ui_action_name",
        ui_action_config_file="ui_action_config_file",
        state_export="state_export",
        ui_action_message="ui_action_message",
        ui_import_loading="ui_import_loading",
    ):
        super().__init__()
        with self:
            with vuetify.VCardText(v_show=(ui_action_name,)):
                vuetify.VBtn(
                    flat=True,
                    icon="mdi-close",
                    style="float: right",
                    click=f"{ui_action_name} = undefined",
                )
                vuetify.VCardTitle(
                    "{{ %s }}" % ui_action_message,
                    v_show=(ui_action_message,),
                )
                with html.Div(v_show=(f"!{ui_action_message}",)):
                    vuetify.VCardTitle("{{ %s }} File Select" % ui_action_name)

                    vuetify.VFileInput(
                        v_model=(ui_action_config_file,),
                        v_show=(f"{ui_action_name} === 'Import'",),
                        accept=".json",
                        label="Config File",
                    )
                    vuetify.VBtn(
                        v_show=(
                            f"{ui_action_name} === 'Import' && {ui_action_config_file} && !{ui_import_loading}",
                        ),
                        variant="tonal",
                        text=(ui_action_name,),
                        click=import_function,
                        style="width: 100%",
                    )
                    vuetify.VCardSubtitle(
                        "Reading configuration file and applying changes...",
                        v_show=(ui_import_loading,),
                    )
                    vuetify.VProgressLinear(
                        v_show=(ui_import_loading,),
                        indeterminate=True,
                    )

                    vuetify.VTextField(
                        v_model=(ui_action_config_file,),
                        v_show=(
                            f"{ui_action_name} === 'Export' && {ui_action_config_file} != false",
                        ),
                        label="Download Location",
                        prepend_icon="mdi-paperclip",
                        # FileSystem API is only available on some browsers:
                        # https://caniuse.com/native-filesystem-api
                        click="""
                            if ($event){
                                try {
                                    window.showSaveFilePicker({
                                        suggestedName: 'pan3d_state.json',
                                        types: [{
                                            accept: {
                                                'application/json': ['.json']
                                            }
                                        }]
                                    }).then((handle) => {
                                        handle.createWritable().then((writable) => {
                                            writable.write(JSON.stringify(%s, null, 4));
                                            writable.close();
                                            %s = 'Export complete.'
                                        })
                                    });
                                } catch {
                                    %s = false;
                                    window.alert('Your browser does not support selecting a download location. Your download will be made to the default location, saved as pan3d_state.json.')
                                }
                            }
                        """
                        % (state_export, ui_action_message, ui_action_config_file),
                    )
                    vuetify.VBtn(
                        v_show=(
                            f"{ui_action_name} === 'Export' && {ui_action_config_file} === false",
                        ),
                        variant="tonal",
                        text="Download pan3d_state.json",
                        click=f"""
                            var content = JSON.stringify({state_export}, null, 4);
                            var a = window.document.createElement('a');
                            a.href = 'data:application/json;charset=utf-8,'  + encodeURIComponent(content);
                            a.download = 'pan3d_state.json';
                            a.style.display = 'none';
                            window.document.body.appendChild(a);
                            a.click()
                            window.document.body.removeChild(a);
                            {ui_action_message} = 'Export complete.'
                        """,
                        style="width: 100%",
                    )



import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleSwitch  # noqa: F401
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

from pan3d.ui.analytics import Plotting
from pan3d.ui.preview import RenderingSettings
from pan3d.ui.vtk_view import Pan3DView
from pan3d.utils.common import ControlPanel, Explorer, SummaryToolbar
from pan3d.utils.convert import to_float
from pan3d.widgets.scalar_bar import ScalarBar
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class Pan3dAnalyticsView(Pan3DView):
    def __init__(self, render_window, **kwargs):
        super().__init__(render_window=render_window, **kwargs)
        with self.toolbar:
            v3.VDivider(classes="my-1")
            with v3.VTooltip(text="Display charting/plotting drawer"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-chart-line",
                        click="plot_drawer = !plot_drawer",
                    )


class AnalyticsExplorer(Explorer):
    def __init__(self, xarray=None, source=None, server=None, local_rendering=None):
        """Create an instance of the AnalyticsExplorer class."""
        super().__init__(xarray, source, server, local_rendering)

        if self.source is None:
            self.source = vtkXArrayRectilinearSource(
                input=self.xarray
            )  # To initialize the pipeline

        self.ui = None
        self._setup_vtk()
        self._build_ui()

    # -------------------------------------------------------------------------
    # VTK Setup
    # -------------------------------------------------------------------------

    def _setup_vtk(self):
        self.renderer = vtkRenderer(background=(0.8, 0.8, 0.8))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        # Need explicit geometry extraction when used with WASM
        self.geometry = vtkDataSetSurfaceFilter(
            input_connection=self.source.output_port
        )
        self.mapper = vtkPolyDataMapper(
            input_connection=self.geometry.output_port,
        )

        self.actor = vtkActor(mapper=self.mapper, visibility=0)

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    # -------------------------------------------------------------------------
    # Trame API
    # -------------------------------------------------------------------------

    def start(self, **kwargs):
        """Initialize the UI and start the server for XArray Viewer."""
        self.ui.server.start(**kwargs)

    @property
    async def ready(self):
        """Start and wait for the XArray Viewer corroutine to be ready."""
        await self.ui.ready

    @property
    def state(self):
        """Returns the current the trame server state."""
        return self.server.state

    @property
    def ctrl(self):
        """Returns the Controller for the trame server."""
        return self.server.controller

    # -------------------------------------------------------------------------
    # UI
    # -------------------------------------------------------------------------

    def _build_ui(self, **kwargs):
        self.state.trame__title = "Analytics Explorer"

        with VAppLayout(self.server, fill_height=True) as layout:
            self.ui = layout
            # Save dialog
            with v3.VDialog(v_model=("show_save_dialog", False)):
                with v3.VCard(classes="mx-auto w-50"):
                    v3.VCardTitle("Save dataset to disk")
                    v3.VDivider()
                    with v3.VCardText():
                        v3.VTextField(
                            label="File path to save",
                            v_model=("save_dataset_path", ""),
                            hide_details=True,
                        )
                    with v3.VCardActions():
                        v3.VSpacer()
                        v3.VBtn(
                            "Save",
                            classes="text-none",
                            variant="flat",
                            color="primary",
                            click=(self.save_dataset, "[save_dataset_path]"),
                        )
                        v3.VBtn(
                            "Cancel",
                            classes="text-none",
                            variant="flat",
                            click="show_save_dialog=false",
                        )

            # Error messages
            v3.VAlert(
                v_if=("data_origin_error", False),
                border="start",
                max_width=700,
                rounded="lg",
                text=("data_origin_error", ""),
                title="Failed to load data",
                type="error",
                variant="tonal",
                style="position:absolute;bottom:1rem;right:1rem;",
            )

            with v3.VLayout():
                with v3.VMain(style="position: relative"):
                    with html.Div(
                        style="position: relative; width: 100%; height: 100%;",
                    ):
                        # 3D view
                        Pan3dAnalyticsView(
                            self.render_window,
                            local_rendering=self.local_rendering,
                            widgets=[self.widget],
                        )

                        # Scalar bar
                        ScalarBar(
                            ctx_name="scalar_bar",
                            v_show="!control_expended",
                            v_if="color_by",
                        )

                        #  # Summary toolbar
                        SummaryToolbar(
                            v_show="!control_expended",
                            v_if="slice_t_max > 0",
                        )

                        # Control panel
                        with ControlPanel(
                            enable_data_selection=(self.xarray is None),
                            toggle="control_expended",
                            load_dataset=self.load_dataset,
                            import_file_upload=self.import_file_upload,
                            export_file_download=self.export_state,
                            xr_update_info="xr_update_info",
                            panel_label="Analytics Explorer",
                        ).ui_content:
                            RenderingSettings(
                                ctx_name="rendering",
                                source=self.source,
                                update_rendering=self.update_rendering,
                            )

                with v3.VNavigationDrawer(
                    disable_resize_watcher=True,
                    disable_route_watcher=True,
                    permanent=True,
                    location="right",
                    v_model=("plot_drawer", False),
                    width=800,
                ):
                    self.plotting = Plotting(
                        source=self.source, toggle="chart_expanded"
                    )

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("color_by")
    def _on_color_by_change_on(self, **kwargs):
        super()._on_color_properties_change(**kwargs)
        self.plotting.update_plot()

    @change("scale_x", "scale_y", "scale_z")
    def _on_scale_change(self, scale_x, scale_y, scale_z, **_):
        self.actor.SetScale(
            to_float(scale_x),
            to_float(scale_y),
            to_float(scale_z),
        )

        if self.state.import_pending:
            return

        if self.actor.visibility:
            self.renderer.ResetCamera()

            if self.local_rendering:
                self.ctrl.view_update(push_camera=True)

            self.ctrl.view_reset_camera()

    def update_rendering(self, reset_camera=False):
        self.state.dirty_data = False

        if self.actor.visibility == 0:
            self.actor.visibility = 1
            self.renderer.AddActor(self.actor)
            self.renderer.ResetCamera()
            if self.ctrl.view_update_force.exists():
                self.ctrl.view_update_force(push_camera=True)

        if reset_camera:
            self.ctrl.view_reset_camera()
        else:
            self.ctrl.view_update()


# -----------------------------------------------------------------------------
# Main executable
# -----------------------------------------------------------------------------


def main():
    app = AnalyticsExplorer()
    app.start()


if __name__ == "__main__":
    main()



from pathlib import Path

import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
import xarray as xr
from vtkmodules.vtkCommonCore import vtkLookupTable
from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkDataSetAttributes
from vtkmodules.vtkFiltersCore import vtkAssignAttribute, vtkCellDataToPointData
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkFiltersModeling import (
    vtkBandedPolyDataContourFilter,
    vtkLoopSubdivisionFilter,
)

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleSwitch  # noqa: F401
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkIOLegacy import vtkDataSetReader
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

from pan3d.ui.css import base, preview
from pan3d.ui.vtk_view import Pan3DView
from pan3d.utils.convert import to_float, to_image
from pan3d.utils.presets import PRESETS, set_preset
from trame.app import get_server
from trame.decorators import TrameApp, change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import html
from trame.widgets import vuetify3 as v3


@TrameApp()
class ContourExplorer:
    def __init__(self, server=None, local_rendering=None):
        self.server = get_server(server, client_type="vue3")
        self.server.enable_module(base)
        self.server.enable_module(preview)

        # CLI
        self.server.cli.add_argument(
            "--wasm",
            help="Use WASM for local rendering",
            action="store_true",
        )
        self.server.cli.add_argument(
            "--vtkjs",
            help="Use vtk.js for local rendering",
            action="store_true",
        )
        self.server.cli.add_argument(
            "--mesh",
            help="Geometry file",
        )
        self.server.cli.add_argument(
            "--fields",
            help="Fields HDF5 file",
        )

        args, _ = self.server.cli.parse_known_args()
        mesh_file = Path(args.mesh).resolve()
        fields_file = Path(args.fields).resolve()

        # Local rendering
        self.local_rendering = local_rendering
        if args.wasm:
            self.local_rendering = "wasm"
        if args.vtkjs:
            self.local_rendering = "vtkjs"

        # setup
        self.last_field = None
        self.last_preset = None
        self.xr = xr.open_dataset(fields_file, engine="netcdf4")
        self.mesh = vtkDataSetReader(file_name=str(mesh_file))()
        self._setup_vtk()
        self._build_ui()

    # -------------------------------------------------------------------------
    # VTK Setup
    # -------------------------------------------------------------------------

    def _setup_vtk(self):
        self.lut = vtkLookupTable()

        self.renderer = vtkRenderer(background=(0.8, 0.8, 0.8))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        # Need explicit geometry extraction when used with WASM
        self.geometry = vtkDataSetSurfaceFilter(input_data=self.mesh)
        self.cell2point = vtkCellDataToPointData(
            input_connection=self.geometry.output_port
        )
        self.refine = vtkLoopSubdivisionFilter(
            input_connection=self.cell2point.output_port, number_of_subdivisions=1
        )
        self.assign = vtkAssignAttribute(input_connection=self.refine.output_port)
        self.assign.Assign(
            "scalars",
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        self.bands = vtkBandedPolyDataContourFilter(
            input_connection=self.assign.output_port,
            generate_contour_edges=1,
        )
        self.mapper = vtkPolyDataMapper(
            input_connection=self.bands.output_port,
            scalar_visibility=1,
            interpolate_scalars_before_mapping=1,
            lookup_table=self.lut,
        )
        self.mapper.SelectColorArray("scalars")
        self.mapper.SetScalarModeToUsePointFieldData()
        self.actor = vtkActor(mapper=self.mapper)

        # contour lines
        self.mapper_lines = vtkPolyDataMapper(
            input_connection=self.bands.GetOutputPort(1),
        )
        self.actor_lines = vtkActor(mapper=self.mapper_lines)
        self.actor_lines.property.color = [0, 0, 0]
        self.actor_lines.property.line_width = 2

        self.renderer.AddActor(self.actor)
        self.renderer.AddActor(self.actor_lines)

        self.renderer.ResetCamera(self.mesh.bounds)

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    # -------------------------------------------------------------------------
    # Trame API
    # -------------------------------------------------------------------------

    @property
    def state(self):
        """Returns the current the trame server state."""
        return self.server.state

    @property
    def ctrl(self):
        """Returns the Controller for the trame server."""
        return self.server.controller

    # -------------------------------------------------------------------------
    # UI
    # -------------------------------------------------------------------------

    def _build_ui(self, **kwargs):
        self.state.update(
            {
                "trame__title": "Contour Explorer",
                "import_pending": False,
                "control_expended": True,
                "axis_names": ["X", "Y", "Z"],
                "scale_x": 1,
                "scale_y": 1,
                "scale_z": 0.01,
            }
        )

        fields = list(self.xr.keys())
        active_field = fields[0]
        nb_times = self.xr[active_field].shape[0]

        with VAppLayout(self.server, fill_height=True) as layout:
            self.ui = layout

            # 3D view
            Pan3DView(
                self.render_window,
                local_rendering=self.local_rendering,
                widgets=[self.widget],
            )

            # Control panel
            with v3.VCard(
                classes="controller", rounded=("control_expended || 'circle'",)
            ):
                with v3.VCardTitle(
                    classes=(
                        "`d-flex pa-1 position-fixed bg-white ${control_expended ? 'controller-content rounded-t border-b-thin':'rounded-circle'}`",
                    ),
                    style="z-index: 1;",
                ):
                    v3.VProgressLinear(
                        v_if=("control_expended", True),
                        indeterminate=("trame__busy",),
                        bg_color="rgba(0,0,0,0)",
                        absolute=True,
                        color="primary",
                        location="bottom",
                        height=2,
                    )
                    v3.VProgressCircular(
                        v_else=True,
                        bg_color="rgba(0,0,0,0)",
                        indeterminate=("trame__busy",),
                        style="position: absolute; top: 0; left: 0; width: 100%; height: 100%;",
                        color="primary",
                        width=3,
                    )
                    v3.VBtn(
                        icon="mdi-close",
                        v_if="control_expended",
                        click="control_expended = !control_expended",
                        flat=True,
                        size="sm",
                    )
                    v3.VBtn(
                        icon="mdi-menu",
                        v_else=True,
                        click="control_expended = !control_expended",
                        flat=True,
                        size="sm",
                    )
                    if self.server.hot_reload:
                        v3.VBtn(
                            v_show="control_expended",
                            icon="mdi-refresh",
                            flat=True,
                            size="sm",
                            click=self.ctrl.on_server_reload,
                        )
                    v3.VSpacer()
                    html.Div(
                        "Contours Explorer",
                        v_show="control_expended",
                        classes="text-h6 px-2",
                    )
                    v3.VSpacer()

                with v3.VCardText(
                    v_show=("control_expended", True),
                    classes="controller-content py-1 mt-10 px-0",
                ):
                    # Actor scaling
                    with v3.VTooltip(text="Representation scaling"):
                        with html.Template(v_slot_activator="{ props }"):
                            with v3.VRow(
                                v_bind="props",
                                no_gutter=True,
                                classes="align-center my-0 mx-0 border-b-thin",
                            ):
                                v3.VIcon(
                                    "mdi-ruler-square",
                                    classes="ml-2 text-medium-emphasis",
                                )
                                with v3.VCol(classes="pa-0", v_if="axis_names?.[0]"):
                                    v3.VTextField(
                                        v_model=("scale_x", 1),
                                        hide_details=True,
                                        density="compact",
                                        flat=True,
                                        variant="solo",
                                        reverse=True,
                                        raw_attrs=[
                                            'pattern="^\d*(\.\d)?$"',
                                            'min="0.001"',
                                            'step="0.1"',
                                        ],
                                        type="number",
                                    )
                                with v3.VCol(classes="pa-0", v_if="axis_names?.[1]"):
                                    v3.VTextField(
                                        v_model=("scale_y", 1),
                                        hide_details=True,
                                        density="compact",
                                        flat=True,
                                        variant="solo",
                                        reverse=True,
                                        raw_attrs=[
                                            'pattern="^\d*(\.\d)?$"',
                                            'min="0.001"',
                                            'step="0.1"',
                                        ],
                                        type="number",
                                    )
                                with v3.VCol(classes="pa-0", v_if="axis_names?.[2]"):
                                    v3.VTextField(
                                        v_model=("scale_z", 1),
                                        hide_details=True,
                                        density="compact",
                                        flat=True,
                                        variant="solo",
                                        reverse=True,
                                        raw_attrs=[
                                            'pattern="^\d*(\.\d)?$"',
                                            'min="0.001"',
                                            'step="0.1"',
                                        ],
                                        type="number",
                                    )

                    v3.VSelect(
                        placeholder="Color By",
                        prepend_inner_icon="mdi-format-color-fill",
                        v_model=("field", fields[0]),
                        items=("fields", fields),
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                    )
                    v3.VDivider()
                    with v3.VRow(no_gutters=True, classes="align-center mr-0"):
                        with v3.VCol():
                            v3.VTextField(
                                prepend_inner_icon="mdi-water-minus",
                                v_model_number=("color_min", 0),
                                type="number",
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                            )
                        with v3.VCol():
                            v3.VTextField(
                                prepend_inner_icon="mdi-water-plus",
                                v_model_number=("color_max", 1),
                                type="number",
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                            )
                        with html.Div(classes="flex-0"):
                            v3.VBtn(
                                icon="mdi-arrow-split-vertical",
                                size="sm",
                                density="compact",
                                flat=True,
                                variant="outlined",
                                classes="mx-2",
                                click=self.reset_color_range,
                            )
                    # v3.VDivider()
                    with html.Div(classes="mx-2"):
                        html.Img(
                            src=("preset_img", None),
                            style="height: 0.75rem; width: 100%;",
                            classes="rounded-lg border-thin",
                        )
                    v3.VSelect(
                        placeholder="Color Preset",
                        prepend_inner_icon="mdi-palette",
                        v_model=("color_preset", "Cool to Warm"),
                        items=("color_presets", list(PRESETS.keys())),
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                    )

                    v3.VDivider()

                    # contours
                    with v3.VTooltip(
                        text=("`Number of contours: ${nb_contours}`",),
                    ):
                        with html.Template(v_slot_activator="{ props }"):
                            with html.Div(
                                classes="d-flex pr-2",
                                v_bind="props",
                            ):
                                v3.VSlider(
                                    v_model=("nb_contours", 20),
                                    min=2,
                                    max=50,
                                    step=1,
                                    prepend_icon="mdi-fingerprint",
                                    hide_details=True,
                                    density="compact",
                                    flat=True,
                                    variant="solo",
                                )

                    v3.VDivider()
                    # Time slider
                    with v3.VTooltip(
                        v_if="slice_t_max > 0",
                        text=("`time: ${time_idx + 1} / ${slice_t_max+1}`",),
                    ):
                        with html.Template(v_slot_activator="{ props }"):
                            with html.Div(
                                classes="d-flex pr-2",
                                v_bind="props",
                            ):
                                v3.VSlider(
                                    prepend_icon="mdi-clock-outline",
                                    v_model=("time_idx", 0),
                                    min=0,
                                    max=("slice_t_max", nb_times - 1),
                                    step=1,
                                    hide_details=True,
                                    density="compact",
                                    flat=True,
                                    variant="solo",
                                )

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("scale_x", "scale_y", "scale_z")
    def _on_scale_change(self, scale_x, scale_y, scale_z, **_):
        self.actor.SetScale(
            to_float(scale_x),
            to_float(scale_y),
            to_float(scale_z),
        )
        self.actor_lines.SetScale(
            to_float(scale_x),
            to_float(scale_y),
            to_float(scale_z),
        )

        if self.actor.visibility:
            self.renderer.ResetCamera()

            if self.local_rendering:
                self.ctrl.view_update(push_camera=True)

            self.ctrl.view_reset_camera()

    @change("field", "time_idx")
    def _on_update_data(self, field, time_idx, **_):
        # update array
        array = self.xr[field][time_idx].values
        self.mesh.cell_data["scalars"] = array

        # update range
        if self.last_field != field:
            self.last_field = field
            self.reset_color_range()

        self.ctrl.view_update()

    @change("color_min", "color_max", "color_preset", "nb_contours")
    def _on_update_color_range(
        self, nb_contours, color_min, color_max, color_preset, **_
    ):
        if self.last_preset != color_preset:
            self.last_preset = color_preset
            set_preset(self.lut, color_preset)
            self.state.preset_img = to_image(self.lut, 255)

        self.mapper.SetScalarRange(color_min, color_max)
        self.bands.GenerateValues(nb_contours, [color_min, color_max])
        self.ctrl.view_update()

    def reset_color_range(self):
        if self.state.field is None:
            return

        field_array = self.xr[self.state.field].values
        with self.state:
            self.state.color_min = float(field_array.min())
            self.state.color_max = float(field_array.max())


def main():
    app = ContourExplorer()
    app.server.start()


if __name__ == "__main__":
    main()



import os
from pathlib import Path

import numpy as np
from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.util import numpy_support
from vtkmodules.vtkCommonDataModel import vtkUniformGrid
from vtkmodules.vtkFiltersCore import vtkAppendDataSets, vtkContourFilter
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkIOImage import vtkJPEGReader
from vtkmodules.vtkIOLegacy import vtkDataSetReader
from vtkmodules.vtkRenderingCore import vtkTexture

from pan3d.filters.globe import ProjectToSphere

DATA_DIR = Path(__file__).with_name("data").resolve()


def get_globe():
    # Add globe underlayment
    # Define grid dimensions (size in each direction)
    grid_dimensions = [361, 181, 1]  # 10x10x10 grid
    uniform_grid = vtkUniformGrid()
    uniform_grid.SetDimensions(grid_dimensions)
    uniform_grid.SetSpacing(1.0, 1.0, 0.0)
    uniform_grid.SetOrigin(-180.0, -90.0, 0.0)

    append = vtkAppendDataSets()
    append.AddInputData(uniform_grid)
    append.Update()

    grid = append.GetOutput()

    grid_np = dsa.WrapDataObject(grid)
    points = np.array(grid_np.Points)
    x = points[:, 0]  # x-coordinate
    y = points[:, 1]  # y-coordinate
    # Map data's longitude (0 to 360) to texture's longitude (-180 to 180)
    x_texture = x - 180  # Convert longitude data to texture's longitude range
    # Normalize the texture coordinates: Longitude (-180 to 180) -> u (0 to 1)
    u = (x_texture) / 360.0
    v = (y + 90.0) / 180.0
    texture_coords_np = np.vstack((u, v)).T
    texture_coords = numpy_support.numpy_to_vtk(texture_coords_np)
    grid.GetPointData().SetTCoords(texture_coords)

    globe = ProjectToSphere()
    globe.input_data_object = grid
    # Need explicit geometry extraction when used with WASM
    geometry = vtkDataSetSurfaceFilter(input_connection=globe.output_port)
    geometry.Update()
    return geometry.output


"""
def get_globe_texture() -> map:
    # Load a texture (JPEG image in this case)
    jpeg_reader = vtkJPEGReader()
    jpeg_reader.SetFileName(
        str(DATA_DIR / "world.topo.bathy.200408.3x5400x2700.jpg")
    )
    jpeg_reader.Update()

    # Create a vtkTexture object
    texture = vtkTexture()
    texture.SetInputConnection(jpeg_reader.GetOutputPort())
    texture.InterpolateOn()

    return texture
"""


def get_globe_textures():
    textures = {}
    textures_dir = DATA_DIR / "globe_textures"
    files = os.listdir(textures_dir)
    for file in files:
        if file.endswith(".md"):
            continue
        texture_name = file[:-4].capitalize()
        texture_path = textures_dir / file

        # Load a texture (JPEG image in this case)
        jpeg_reader = vtkJPEGReader()
        jpeg_reader.SetFileName(str(texture_path))
        jpeg_reader.Update()

        # Create a vtkTexture object
        texture = vtkTexture()
        texture.SetInputConnection(jpeg_reader.GetOutputPort())
        texture.InterpolateOn()

        textures[texture_name] = texture

    return textures


def get_continent_outlines():
    vtk_reader = vtkDataSetReader()
    vtk_reader.SetFileName(str(DATA_DIR / "continents.vtk"))
    vtk_reader.Update()
    vtk_reader.output.GetPointData().SetActiveScalars("cstar")

    contour = vtkContourFilter()
    contour.input_data_object = vtk_reader.output
    contour.SetValue(0, 0.5)
    contour.SetNumberOfContours(1)
    contour.Update()

    continents = ProjectToSphere()
    continents.input_data_object = contour.output
    continents.Update()
    # Need explicit geometry extraction when used with WASM
    geometry = vtkDataSetSurfaceFilter(input_connection=continents.output_port)
    geometry.Update()

    return geometry.output



import importlib


def _call_catalog_function(catalog_name, function_name, **kwargs):
    try:
        module = importlib.import_module(f"pan3d.catalogs.{catalog_name}")
        func = getattr(module, function_name)
        return func(**kwargs)
    except ImportError as in_err:
        import_msg = f"{catalog_name} catalog module not enabled. Install pan3d[{catalog_name}] to load this catalog."
        raise ValueError(import_msg) from in_err
    except AttributeError as att_err:
        attr_msg = f"{catalog_name} is not a valid catalog module."
        raise ValueError(attr_msg) from att_err


def get(catalog_name):
    return _call_catalog_function(catalog_name, "get_catalog")


def get_search_options(catalog_name):
    return _call_catalog_function(catalog_name, "get_search_options")


def search(catalog_name, **filters):
    return _call_catalog_function(catalog_name, "search", **filters)


def load_dataset(catalog_name, id):
    return _call_catalog_function(catalog_name, "load_dataset", id=id)


def list_availables():
    # FIXME - make it dynamic
    return [
        {"value": "file", "title": "Local file"},
        {"value": "url", "title": "Remote URL"},
        {"value": "xarray", "title": "XArray Tutorial"},
        # {"value": "esgf", "title": "ESGF"},
        # {"value": "pangeo", "title": "Pangeo"},
    ]


def list_availables_search():
    # FIXME do a try import and if empty suggest pip install
    return [
        {"id": "esgf", "name": "ESGF", "search_terms": [{"key": "id", "options": []}]},
        {
            "id": "pangeo",
            "name": "Pangeo Forge",
            "search_terms": [{"key": "id", "options": []}],
        },
    ]


__all__ = [
    "get",
    "get_search_options",
    "list_availables",
    "load_dataset",
    "search",
]



import os


def has_gpu():
    # Detect known environments without gpu rendering
    target_env_vars = ["BINDER_REQUEST"]
    return not any(os.environ.get(k) for k in target_env_vars)


XARRAY_EXAMPLES = [
    # from https://docs.xarray.dev/en/stable/generated/xarray.tutorial.open_dataset.html
    {
        "name": "XArray Examples - Air Temperature",
        "value": {"source": "xarray", "id": "air_temperature"},
    },
    {
        "name": "XArray Examples - Ocean Basins",
        "value": {"source": "xarray", "id": "basin_mask"},
    },
    {
        "name": "XArray Examples - Ice Velocity",
        "value": {"source": "xarray", "id": "ASE_ice_velocity"},
    },
    {
        "name": "XArray Examples - Regional Arctic System Model",
        "value": {"source": "xarray", "id": "rasm"},
    },
    {
        "name": "XArray Examples - Regional Ocean Model System",
        "value": {"source": "xarray", "id": "ROMS_example"},
    },
    {
        "name": "XArray Examples - ERA-Interim analysis",
        "value": {"source": "xarray", "id": "eraint_uvz"},
    },
    {
        "name": "XArray Examples - NOAA Sea Surface Temperatures",
        "value": {"source": "xarray", "id": "ersstv5"},
    },
]

# Used in preview

XYZ = ["x", "y", "z"]
SLICE_VARS = ["slice_{}_range", "slice_{}_cut", "slice_{}_type", "slice_{}_step"]
VIEW_UPS = {
    (1, 1, 1): (0, 0, 1),
    (-1, -1, 1): (0, 0, 1),
    (1, 0, 0): (0, 0, 1),
    (0, 1, 0): (0, 0, 1),
    (0, 0, 1): (0, 1, 0),
}



from trame_client.widgets.core import AbstractElement

from trame.widgets import vuetify3 as v3


class CollapsableSection(AbstractElement):
    id_count = 0

    def __init__(self, title, var_name=None, expended=False, **kwargs):
        super().__init__(None, **kwargs)
        CollapsableSection.id_count += 1
        show = var_name or f"show_section_{CollapsableSection.id_count}"
        with v3.VCardSubtitle(
            classes="px-0 ml-n3 d-flex align-center font-weight-bold pointer",
            click=f"{show} = !{show}",
        ) as container:
            v3.VIcon(
                f"{{{{ {show} ? 'mdi-menu-down' : 'mdi-menu-right' }}}}",
                size="sm",
                classes="pa-0 ma-0",
            )
            container.add_child(title)
        self.content = v3.VSheet(
            border="opacity-25 thin",
            classes="overflow-hidden mx-auto mt-1 mb-2",
            rounded="lg",
            v_show=(show, expended),
        )



from typing import Optional

import numpy as np
from vtkmodules.vtkCommonCore import vtkLookupTable

from pan3d.utils.convert import to_image
from pan3d.utils.presets import PRESETS, set_preset
from trame.widgets import html
from trame.widgets import vuetify3 as v3

POINT_DATA = "point_data"
CELL_DATA = "cell_data"
FIELD_DATA = "field_data"


class ColorBy(html.Div):
    """
    Color settings for the XArray Explorers.
    """

    _next_id = 0

    @classmethod
    def next_id(cls):
        """Get the next unique ID for the scalar bar."""
        cls._next_id += 1
        return f"pan3d_scalarbar{cls._next_id}"

    def __init__(
        self,
        data_arrays: Optional[list[dict]] = None,
        color_by=None,
        preset="Fast",
        color_min=0.0,
        color_max=1.0,
        nan_color=0,
        color_by_name=None,
        preset_name=None,
        color_min_name=None,
        color_max_name=None,
        nan_color_name=None,
        reset_color_range=None,
        **kwargs,
    ):
        """
        Initialize the ColorBy UI component.

        Parameters
        ----------
        data_arrays : List[Dict], optional
            A list of dictionaries representing available arrays to color by.
            Each dictionary should contain keys such for name, association, and array min and max.
            The association can be 'point_data', 'cell_data', or 'field_data'
            e.g. [{'name' : 'temperature', 'assoc' : 'point_data', 'min' : 0.0, 'max' : 100.0}]

        color_by : str or None, optional
            The name of the currently selected array to color by.

        preset : str, optional
            Name of the colormap preset to use. Defaults to "Fast".

        color_min : float, optional
            Minimum value of the color mapping range. Defaults to 0.0.

        color_max : float, optional
            Maximum value of the color mapping range. Defaults to 1.0.

        nan_color : int or tuple, optional
            Color to use for NaN values (index into colormap). Defaults to 0.

        color_by_name : str or None, optional
            Name of the UI variable to bind the `color_by` selection to.

        preset_name : str or None, optional
            Name of the UI variable to bind the `preset` selection to.

        color_min_name : str or None, optional
            Name of the UI variable to bind the `color_min` value to.

        color_max_name : str or None, optional
            Name of the UI variable to bind the `color_max` value to.

        nan_color_name : str or None, optional
            Name of the UI variable to bind the `nan_color` value to.

        reset_color_range : callable or None, optional
            Optional callback function to reset the color range to the array's min/max.

        **kwargs : dict
            Additional keyword arguments passed to the parent `html.Div` component.
        """

        self._lut = vtkLookupTable()
        super().__init__(**kwargs)

        ns = self.next_id()
        # Variables that serve and input/output (interactive) can be user specified
        self.__color_by = color_by_name or f"{ns}_color_by"
        self.__color_preset = preset_name or f"{ns}_preset"
        self.__color_min = color_min_name or f"{ns}_color_min"
        self.__color_max = color_max_name or f"{ns}_color_max"
        self.__nan_color = nan_color_name or f"{ns}_nan_color"

        # Variables that are only input or only output do not need user specification
        self.__data_arrays = f"{ns}_data_arrays"
        self.__preset_image = f"{ns}_preset_img"
        self.__color_presets = f"{ns}_color_presets"

        # Register changes based on state update within the widget
        self.state.change(self.__color_preset)(self.set_preset)
        self.state.change(self.__color_by)(self.set_color_by)
        self.state.change(self.__nan_color)(self.set_nan_color)

        self.__array_infos = None
        self.data_arrays = data_arrays
        self.color_by = color_by
        self.preset = preset
        self.color_min = color_min
        self.color_max = color_max
        self.nan_color = nan_color

        with self:
            v3.VSelect(
                placeholder="Color By",
                prepend_inner_icon="mdi-format-color-fill",
                v_model=(self.__color_by, None),
                items=(self.__data_arrays, []),
                clearable=True,
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
            )
            v3.VDivider()
            with v3.VRow(no_gutters=True, classes="align-center mr-0"):
                with v3.VCol():
                    v3.VTextField(
                        prepend_inner_icon="mdi-water-minus",
                        v_model_number=(self.__color_min, 0.45),
                        type="number",
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                        reverse=True,
                    )
                with v3.VCol():
                    v3.VTextField(
                        prepend_inner_icon="mdi-water-plus",
                        v_model_number=(self.__color_max, 5.45),
                        type="number",
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                        reverse=True,
                    )
                with html.Div(classes="flex-0"):
                    v3.VBtn(
                        icon="mdi-arrow-split-vertical",
                        size="sm",
                        density="compact",
                        flat=True,
                        variant="outlined",
                        classes="mx-2",
                        click=reset_color_range,
                    )
            # v3.VDivider()
            with html.Div(classes="mx-2"):
                html.Img(
                    src=(self.__preset_image, None),
                    style="height: 0.75rem; width: 100%;",
                    classes="rounded-lg border-thin",
                )
            v3.VSelect(
                placeholder="Color Preset",
                prepend_inner_icon="mdi-palette",
                v_model=(self.__color_preset, "Fast"),
                items=(self.__color_presets, list(PRESETS.keys())),
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
            )

            with v3.VTooltip(
                text=("`NaN Color (${nan_colors[nan_color]})`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with v3.VItemGroup(
                        v_model=self.__nan_color,
                        v_bind="props",
                        classes="d-inline-flex ga-1 pa-2",
                        mandatory="force",
                    ):
                        v3.VIcon(
                            "mdi-eyedropper-variant",
                            classes="my-auto mx-1 text-medium-emphasis",
                        )
                        with v3.VItem(
                            v_for="(color, i) in nan_colors", key="i", value=("i",)
                        ):
                            with v3.Template(
                                raw_attrs=['#default="{ isSelected, toggle }"']
                            ):
                                with v3.VAvatar(
                                    density="compact",
                                    color=("isSelected ? 'primary': 'transparent'",),
                                ):
                                    v3.VBtn(
                                        "{{ color[3] < 0.1 ? 't' : '' }}",
                                        density="compact",
                                        border="md surface opacity-100",
                                        color=(
                                            "color[3] ? `rgb(${color[0] * 255}, ${color[1] * 255}, ${color[2] * 255})` : undefined",
                                        ),
                                        flat=True,
                                        icon=True,
                                        ripple=False,
                                        size="small",
                                        click="toggle",
                                    )

    @property
    def data_arrays(self):
        """
        Returns the arrays available to color the data by based on the list of dictionaries representing array metadata
        Each dictionary contains keys such for name, association, and array min and max.
        The association can be either of 'point_data', 'cell_data', or 'field_data'
        e.g. [{'name' : 'temperature', 'assoc' : 'point_data', 'min' : 0.0, 'max' : 100.0}]
        """
        return self.__array_infos

    @data_arrays.setter
    def data_arrays(self, array_info: list[dict]):
        """
        Controls the arrays available to color the data by based on the list of dictionaries representing array metadata
        Each dictionary should contain keys such for name, association, and array min and max.
        The association can be either of 'point_data', 'cell_data', or 'field_data'
        e.g. [{'name' : 'temperature', 'assoc' : 'point_data', 'min' : 0.0, 'max' : 100.0}]
        """
        if array_info is None:
            return
        self.__array_infos = array_info
        data_arrays = [info["name"] for info in array_info]
        self.state[self.__data_arrays] = data_arrays

        color_by = self.color_by
        # If the data arrays are empty, set color_by to None
        if array_info is None or len(data_arrays) == 0:
            self.color_by = None
        # If the color_by is not in the new data arrays, reset it to the first available
        elif color_by is None or color_by not in data_arrays:
            self.color_by = data_arrays[0]

    @property
    def color_by(self):
        return self.state[self.__color_by]

    def set_color_by(self, **kwargs):
        self.color_by = self.state[self.__color_by]

    @color_by.setter
    def color_by(self, value):
        """Set the array to color by."""
        with self.state:
            self.state[self.__color_by] = value
            if self.__array_infos:
                info = next(
                    (info for info in self.__array_infos if info["name"] == value), None
                )
                if info is not None:
                    self.color_min = float(info.get("min", self.color_min))
                    self.color_max = float(info.get("max", self.color_max))
                else:
                    self.color_min = 0.0
                    self.color_max = 1.0

    @property
    def color_by_name(self):
        return self.__color_by

    def set_preset(self, **_):
        """Set the color preset for the scalar bar."""
        self.preset = self.state[self.__color_preset]

    @property
    def preset(self):
        return self._preset

    @preset.setter
    def preset(self, value):
        """Set the color preset to color the data by."""
        if value not in PRESETS:
            err_msg = f"Preset '{value}' not found."
            raise ValueError(err_msg)
        self._preset = value
        set_preset(self._lut, value)
        with self.state:
            self.state[self.__preset_image] = to_image(self._lut)

    @property
    def preset_image_name(self):
        return self.__preset_image

    def set_color_range(self, color_min, color_max):
        """Set the color range for the scalar bar."""
        self.state[self.__color_min] = color_min
        self.state[self.__color_max] = color_max

    @property
    def color_min(self):
        return self.state[self.__color_min]

    @color_min.setter
    def color_min(self, value):
        """Set the minimum value of the color mapping range."""
        with self.state:
            self.state[self.__color_min] = value

    @property
    def color_min_name(self):
        return self.__color_min

    @property
    def color_max(self):
        return self.state[self.__color_max]

    @color_max.setter
    def color_max(self, value):
        """Set the maximum value of the color mapping range."""
        with self.state:
            self.state[self.__color_max] = value

    @property
    def color_max_name(self):
        return self.__color_max

    @property
    def nan_color(self):
        nan_colors = self.state.nan_colors
        nan_color = self.state[self.__nan_color]
        return nan_colors.get(nan_color, [0, 0, 0, 0])

    @nan_color.setter
    def nan_color(self, value: int):
        """Set the color for NaN values."""
        with self.state:
            self.state[self.__nan_color] = value
            nan_color = self.state.nan_colors[value]
            self._lut.SetNanColor(nan_color)

    def set_nan_color(self, **_):
        nan_colors = self.state.nan_colors
        nan_color = nan_colors[self.state[self.__nan_color]]
        self._lut.SetNanColor(nan_color)

    def set_data_arrays_from_vtk(self, dataset, associations=None):
        """Inspect dataset to extract arrays metadata. Only works with scalar fields."""
        if dataset is None:
            self.data_arrays = []
            return

        array_info = []
        if associations is None:
            associations = [POINT_DATA, CELL_DATA, FIELD_DATA]

        for association in associations:
            arrays = getattr(dataset, association, None)
            if arrays is not None:
                for array in arrays:
                    array_info.append(
                        {
                            "name": array.VTKObject.GetName(),
                            "min": np.min(array),
                            "max": np.max(array),
                            "assoc": association,
                        }
                    )
        self.data_arrays = array_info

    def configure_mapper(self, mapper):
        """Configure the color mapper with the current settings for any data association."""
        # Find the association type for the selected array
        assoc = None
        if self.__array_infos:
            info = next(
                (info for info in self.__array_infos if info["name"] == self.color_by),
                None,
            )
            if info is not None:
                assoc = info.get("assoc", POINT_DATA).lower()
            else:
                mapper.SetScalarVisibility(0)
                return

        # Set the color array and scalar mode based on association
        if assoc == POINT_DATA:
            mapper.SetScalarModeToUsePointFieldData()
        elif assoc == CELL_DATA:
            mapper.SetScalarModeToUseCellFieldData()
        elif assoc == FIELD_DATA:
            mapper.SetScalarModeToUseFieldData()

        set_preset(self._lut, self.preset)
        mapper.SetLookupTable(self._lut)
        mapper.SelectColorArray(self.color_by)
        mapper.SetScalarRange(self.color_min, self.color_max)
        mapper.SetScalarVisibility(1)



import math

from pan3d.utils.common import RenderingSettingsBasic
from pan3d.utils.convert import max_str_length
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class ContourRenderingSettings(RenderingSettingsBasic):
    def __init__(self, source, update_rendering, **kwargs):
        super().__init__(source, update_rendering, **kwargs)

        self.source = source

        with self.content:
            # Actor scaling
            with v3.VTooltip(text="Representation scaling"):
                with html.Template(v_slot_activator="{ props }"):
                    with v3.VRow(
                        v_bind="props",
                        no_gutter=True,
                        classes="align-center my-0 mx-0 border-b-thin",
                    ):
                        v3.VIcon(
                            "mdi-ruler-square",
                            classes="ml-2 text-medium-emphasis",
                        )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[0]"):
                            v3.VTextField(
                                v_model=("scale_x", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[1]"):
                            v3.VTextField(
                                v_model=("scale_y", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )
                        with v3.VCol(classes="pa-0", v_if="axis_names?.[2]"):
                            v3.VTextField(
                                v_model=("scale_z", 1),
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                reverse=True,
                                raw_attrs=[
                                    'pattern="^\d*(\.\d)?$"',
                                    'min="0.001"',
                                    'step="0.1"',
                                ],
                                type="number",
                            )

            # contours
            with v3.VTooltip(
                text=("`Number of contours: ${nb_contours}`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            v_model=("nb_contours", 20),
                            min=2,
                            max=50,
                            step=1,
                            prepend_icon="mdi-fingerprint",
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )

            v3.VDivider()
            # Time slider
            with v3.VTooltip(
                v_if="slice_t_max > 0",
                text=("`time: ${time_idx + 1} / ${slice_t_max+1}`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    with html.Div(
                        classes="d-flex pr-2",
                        v_bind="props",
                    ):
                        v3.VSlider(
                            prepend_icon="mdi-clock-outline",
                            v_model=("slice_t", 0),
                            min=0,
                            max=("slice_t_max", 0),
                            step=1,
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                        )
            v3.VDivider()
            v3.VBtn(
                "Update 3D view",
                block=True,
                classes="text-none",
                flat=True,
                density="compact",
                rounded=0,
                disabled=("data_arrays.length === 0",),
                color=("dirty_data && data_arrays.length ? 'primary': undefined",),
                click=(update_rendering, "[true]"),
            )

    def update_from_source(self, source=None):
        if source is None:
            return

        with self.state as state:
            state.data_arrays_available = source.available_arrays
            state.data_arrays = source.arrays
            state.color_by = None
            state.axis_names = [source.x, source.y, source.z]
            state.slice_extents = source.slice_extents

            # Update time
            state.slice_t = source.t_index
            state.slice_t_max = source.t_size - 1
            state.t_labels = source.t_labels
            state.max_time_width = math.ceil(0.58 * max_str_length(state.t_labels))

            if state.slice_t_max > 0:
                state.max_time_index_width = math.ceil(
                    0.6 + (math.log10(state.slice_t_max + 1) + 1) * 2 * 0.58
                )
